# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records API."""

from functools import partial

import pytest

from invenio_records.api import Record
from invenio_records.signals import (
    after_record_delete,
    after_record_insert,
    after_record_revert,
    after_record_update,
    before_record_delete,
    before_record_insert,
    before_record_revert,
    before_record_update,
)


@pytest.fixture()
def signals():
    """Fixtures to connect signals."""
    called = {}

    def _listener(signal_name, sender, *args, **kwargs):
        if signal_name not in called:
            called[signal_name] = 0
        called[signal_name] += 1

    after_record_delete_listener = partial(_listener, "after_record_delete")
    after_record_insert_listener = partial(_listener, "after_record_insert")
    after_record_revert_listener = partial(_listener, "after_record_revert")
    after_record_update_listener = partial(_listener, "after_record_update")
    before_record_delete_listener = partial(_listener, "before_record_delete")
    before_record_insert_listener = partial(_listener, "before_record_insert")
    before_record_revert_listener = partial(_listener, "before_record_revert")
    before_record_update_listener = partial(_listener, "before_record_update")
    before_record_insert_listener = partial(_listener, "before_record_insert")

    after_record_delete.connect(after_record_delete_listener)
    after_record_insert.connect(after_record_insert_listener)
    after_record_revert.connect(after_record_revert_listener)
    after_record_update.connect(after_record_update_listener)
    before_record_delete.connect(before_record_delete_listener)
    before_record_insert.connect(before_record_insert_listener)
    before_record_revert.connect(before_record_revert_listener)
    before_record_update.connect(before_record_update_listener)
    before_record_insert.connect(before_record_insert_listener)

    yield called

    after_record_delete.disconnect(after_record_delete_listener)
    after_record_insert.disconnect(after_record_insert_listener)
    after_record_revert.disconnect(after_record_revert_listener)
    after_record_update.disconnect(after_record_update_listener)
    before_record_delete.disconnect(before_record_delete_listener)
    before_record_insert.disconnect(before_record_insert_listener)
    before_record_revert.disconnect(before_record_revert_listener)
    before_record_update.disconnect(before_record_update_listener)
    before_record_insert.disconnect(before_record_insert_listener)


def test_signals(testapp, database, signals):
    """Test signals being sent."""
    db = database
    record = Record.create({"title": "Test"})
    db.session.commit()
    assert "before_record_insert" in signals
    assert "after_record_insert" in signals
    assert len(signals.keys()) == 2

    record["title"] = "Test2"
    record.commit()
    db.session.commit()
    assert "before_record_update" in signals
    assert "after_record_update" in signals
    assert len(signals.keys()) == 4

    record.revert(0)
    db.session.commit()
    assert "before_record_revert" in signals
    assert "after_record_revert" in signals
    assert len(signals.keys()) == 6

    record.delete()
    db.session.commit()
    assert "before_record_delete" in signals
    assert "after_record_delete" in signals
    assert len(signals.keys()) == 8


def test_signals_disabled(testapp, database, signals):
    """Test signals being sent."""
    db = database

    class MyRecord(Record):
        # Disable signals
        send_signals = False

    # Same operations as above (but none of them should send signals)
    record = MyRecord.create({"title": "Test"})
    db.session.commit()
    record["title"] = "Test2"
    record.commit()
    db.session.commit()
    record.revert(0)
    db.session.commit()
    record.delete()
    db.session.commit()

    # Assert that no signals was sent.
    assert len(signals.keys()) == 0



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Dumpers used for producing versions of records that can be persisted.

**A simple example**

You can for instance use a dumper to produce the body of the document to be
indexed for the search engine:

.. code-block:: python

    dump = Record({...}).dumps(dumper=SearchDumper())

A dump can be loaded by the dumper as well:

.. code-block:: python

    record = Record.loads(dump, loader=SearchDumper())

**Data harmonization**

Invenio can read records from the database, search engine and data files. The
master copy is always the database, however for performance reasons, it's not
efficient to always use the master version of a record. For instance, during
searches it would come with a big performance impact if we had to read each
record in a search result from the database.

The problem is however that a secondary copy of a record (e.g. in the
search index) is not identical to the master copy. For instance, usage
statistics may have been cached in the search engine version whereas we don't
persist it in the database. This is again for performance reasons and allows
e.g. also having a "sort by most viewed" while not overloading the database
with usage statistics updates.

Because the master and secondary copies might not be identical, this causes
troubles for other Invenio modules who would have to "massage" the record
depending on where it comes form. Overall, this eventually leads to a confusing
data flow in the application.

The dumpers fixes this issue by harmonizing data access to a record from
multiple different data sources. This way, other Invenio modules always have a
standardized version of a record independently of where it was loaded from.
"""

from .base import Dumper
from .search import SearchDumper, SearchDumperExt

__all__ = (
    "Dumper",
    "SearchDumper",
    "SearchDumperExt",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records."""


import uuid

import pytest
from invenio_db import InvenioDB
from invenio_db import db as db_
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy_utils.functions import create_database, database_exists

from invenio_records import Record
from invenio_records.models import RecordMetadataBase


@pytest.fixture(scope="module")
def CustomMetadata(database):
    """."""
    db = database

    class CustomMetadata(db.Model, RecordMetadataBase):
        __tablename__ = "custom_metadata"

    CustomMetadata.__table__.create(db.engine)
    return CustomMetadata


def test_class_model(testapp, database, CustomMetadata):
    """Test custom class model."""
    db = database

    class CustomRecord(Record):
        model_cls = CustomMetadata

    assert "custom_metadata" in db.metadata.tables.keys()

    recid = uuid.UUID("262d2748-ba41-456f-a844-4d043a419a6f")

    # Create a new record with two mutables, a list and a dict
    rec = CustomRecord.create(
        {
            "title": "Title",
            "list": [
                "foo",
            ],
            "dict": {"moo": "boo"},
        },
        id_=recid,
    )

    db.session.commit()
    db.session.expunge_all()
    # record should be in the table
    rec = CustomRecord.get_record(recid)
    assert rec == {
        "title": "Title",
        "list": [
            "foo",
        ],
        "dict": {"moo": "boo"},
    }
    # the record should not be in the default table
    pytest.raises(NoResultFound, Record.get_record, recid)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.


"""Pytest configuration."""

import pytest
from flask import Flask
from invenio_celery import InvenioCelery
from invenio_db import InvenioDB
from invenio_i18n import InvenioI18N
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.schema import DropConstraint, DropSequence, DropTable

from invenio_records import InvenioRecords
from invenio_records.api import Record
from invenio_records.systemfields import SystemFieldsMixin

pytest_plugins = ("celery.contrib.pytest",)


@compiles(DropTable, "postgresql")
def _compile_drop_table(element, compiler, **kwargs):
    return compiler.visit_drop_table(element) + " CASCADE"


@compiles(DropConstraint, "postgresql")
def _compile_drop_constraint(element, compiler, **kwargs):
    return compiler.visit_drop_constraint(element) + " CASCADE"


@compiles(DropSequence, "postgresql")
def _compile_drop_sequence(element, compiler, **kwargs):
    return compiler.visit_drop_sequence(element) + " CASCADE"


@pytest.fixture(scope="module")
def create_app(instance_path):
    """Application factory fixture for use with pytest-invenio."""

    def _create_app(**config):
        app_ = Flask(
            __name__,
            instance_path=instance_path,
        )
        app_.config.update(config)
        InvenioCelery(app_)
        InvenioDB(app_)
        InvenioRecords(app_)
        InvenioI18N(app_)
        return app_

    return _create_app


@pytest.fixture(scope="module")
def testapp(base_app, database):
    """Application with just a database.

    Pytest-Invenio also initialises ES with the app fixture.
    """
    yield base_app


@pytest.fixture()
def languages(db):
    """Languages fixture."""

    class Language(Record, SystemFieldsMixin):
        pass

    languages_data = (
        {
            "title": "English",
            "iso": "en",
            "information": {"native_speakers": "400 million", "ethnicity": "English"},
        },
        {
            "title": "French",
            "iso": "fr",
            "information": {"native_speakers": "76.8 million", "ethnicity": "French"},
        },
        {
            "title": "Spanish",
            "iso": "es",
            "information": {"native_speakers": "489 million", "ethnicity": "Spanish"},
        },
        {
            "title": "Italian",
            "iso": "it",
            "information": {"native_speakers": "67 million", "ethnicity": "Italians"},
        },
        {
            "title": "Old English",
            "iso": "oe",
            "information": {
                "native_speakers": "400 million",
                "ethnicity": ["English", "Old english"],
            },
        },
    )

    languages = {}
    for lang in languages_data:
        lang_rec = Language.create(lang)
        languages[lang["iso"]] = lang_rec
    db.session.commit()
    return Language, languages



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Indexed timestamp dumper.

Dumper used to dump/load the indexed time of a record to/from a search engine body.
"""

import arrow

from .search import SearchDumperExt


class IndexedAtDumperExt(SearchDumperExt):
    """Dumper for the indexed_at field."""

    def __init__(self, key="indexed_at"):
        """Initialize the dumper."""
        self.key = key

    def dump(self, record, data):
        """Dump relations."""
        data[self.key] = arrow.utcnow().isoformat()

    def load(self, data, record_cls):
        """Load (remove) indexed data."""
        data.pop(self.key, None)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Errors for Invenio-Records module."""


class RecordsError(Exception):
    """Base class for errors in Invenio-Records module."""


class MissingModelError(RecordsError):
    """Error raised when a record has no model."""


class RecordsRefResolverConfigError(RecordsError):
    """Custom ref resolver configuration it not correct."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Test JSON resolver."""


import jsonresolver


@jsonresolver.route("/<item>", host="nest.ed")
def test_resolver(item):
    """Create a nested JSON."""
    next_ = (
        {
            "$ref": "http://nest.ed/{}".format(item[1:]),
        }
        if len(item[1:])
        else "."
    )
    return {"letter": item[0], "next": next_}



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field.

The field consists of:

- The system field itself.
- A relation mapping which is an object returned when accessing the system
  field from an instance (e.g. ``record.relations``).
- Relations (e.g. PKRelation) used for defining a relationship to a specific
  record type.
- Results, values returned when accessing a field (e.g.
  ``record.relations.languages``).
"""

from .errors import InvalidCheckValue, InvalidRelationValue, RelationError
from .field import MultiRelationsField, RelationsField
from .mapping import RelationsMapping
from .modelrelations import ModelRelation
from .relations import (
    ListRelation,
    NestedListRelation,
    PKListRelation,
    PKNestedListRelation,
    PKRelation,
    RelationBase,
)
from .results import RelationListResult, RelationNestedListResult, RelationResult

__all__ = (
    "InvalidCheckValue",
    "InvalidRelationValue",
    "ListRelation",
    "ModelRelation",
    "MultiRelationsField",
    "NestedListRelation",
    "PKListRelation",
    "PKNestedListRelation",
    "PKRelation",
    "RelationBase",
    "RelationError",
    "RelationListResult",
    "RelationNestedListResult",
    "RelationResult",
    "RelationsField",
    "RelationsMapping",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Relations fields."""

from collections.abc import Iterable

import pytest

from invenio_records.api import Record
from invenio_records.systemfields import (
    MultiRelationsField,
    RelationsField,
    SystemFieldsMixin,
)
from invenio_records.systemfields.relations import (
    InvalidRelationValue,
    PKListRelation,
    RelationsMapping,
)


def test_multirelations_field(testapp, db, languages):
    """MultiRelationsField tests."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]

    def _assert_iter_and_lang_from_record(record):
        res_iter = record.relations.languages()
        assert isinstance(res_iter, Iterable)
        res = next(res_iter)
        assert res == en_lang

        res_iter = record.relations.inner_languages()
        assert isinstance(res_iter, Iterable)
        res = next(res_iter)
        assert res == fr_lang

    class Record1(Record, SystemFieldsMixin):
        relations = MultiRelationsField(
            languages=PKListRelation(
                key="languages",
                keys=["iso", "information.ethnicity"],
                record_cls=Language,
            ),
            inner=RelationsField(
                inner_languages=PKListRelation(
                    key="inner.inner_languages",
                    keys=["iso", "information.ethnicity"],
                    record_cls=Language,
                ),
            ),
        )

    # Class-field check
    # ---------------------------------------------------------------
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.languages, PKListRelation)
    assert isinstance(Record1.relations.inner_languages, PKListRelation)
    with pytest.raises(AttributeError):
        Record1.inner  # keys are flattened, accessing this field gives an error

    # Empty initialization
    # ---------------------------------------------------------------
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.languages)
    assert callable(record.relations.inner_languages)
    assert record.relations.languages() is None
    assert record.relations.inner_languages() is None

    # Initialize from dictionary data
    # ---------------------------------------------------------------
    record = Record1(
        {
            "languages": [{"id": str(en_lang.id)}],
            "inner": {
                "inner_languages": [{"id": str(fr_lang.id)}],
            },
        }
    )
    _assert_iter_and_lang_from_record(record)

    # Initialize from field data
    # ---------------------------------------------------------------
    record = Record1(
        {},
        relations={
            "languages": [str(en_lang.id)],
            "inner": {
                "inner_languages": [str(fr_lang.id)],
            },
        },
    )
    assert record["languages"] == [{"id": str(en_lang.id)}]
    assert record["inner"]["inner_languages"] == [{"id": str(fr_lang.id)}]
    _assert_iter_and_lang_from_record(record)

    # Set via record dictionary data
    # ---------------------------------------------------------------
    record = Record1.create({})
    record["languages"] = [{"id": str(en_lang.id)}]
    record["inner"] = {}
    record["inner"]["inner_languages"] = [{"id": str(fr_lang.id)}]
    _assert_iter_and_lang_from_record(record)
    record.commit()  # validates
    _assert_iter_and_lang_from_record(record)

    record["languages"] = {"id": "invalid"}
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation
    record["languages"] = [{"id": str(en_lang.id)}]
    record["inner"]["inner_languages"] = {"id": "invalid"}
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    # ---------------------------------------------------------------
    record = Record1.create({})
    record.relations.languages = [str(en_lang.id)]
    record.relations.inner_languages = [str(fr_lang.id)]

    assert record["languages"][0]["id"] == str(en_lang.id)
    assert record["inner"]["inner_languages"][0]["id"] == str(fr_lang.id)
    _assert_iter_and_lang_from_record(record)

    # Update existing value
    # ---------------------------------------------------------------
    record.relations.languages = [str(fr_lang.id)]
    record.relations.inner_languages = [str(en_lang.id)]
    assert record["languages"][0]["id"] == str(fr_lang.id)
    assert record["inner"]["inner_languages"][0]["id"] == str(en_lang.id)

    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    res_iter = record.relations.inner_languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Set invalid value
    # ---------------------------------------------------------------
    record = Record1.create({})
    record.relations.languages = [str(en_lang.id)]
    record.relations.inner_languages = [str(fr_lang.id)]

    for v in ("invalid", ["invalid"]):
        with pytest.raises(InvalidRelationValue):
            record.relations.languages = v

        with pytest.raises(InvalidRelationValue):
            record.relations.inner_languages = v
        # Check that old value is still there
        assert record["languages"][0]["id"] == str(en_lang.id)
        assert record["inner"]["inner_languages"][0]["id"] == str(fr_lang.id)
        _assert_iter_and_lang_from_record(record)

    # Dereference relation
    # ---------------------------------------------------------------
    record.relations.languages.dereference()
    assert record["languages"] == [
        {
            "id": str(en_lang.id),
            "iso": "en",  # only stores the "iso" field
            "information": {"ethnicity": "English"},
            "@v": str(en_lang.id) + "::" + str(en_lang.revision_id),
        }
    ]
    record.relations.inner_languages.dereference()
    assert record["inner"]["inner_languages"] == [
        {
            "id": str(fr_lang.id),
            "iso": "fr",  # only stores the "iso" field
            "information": {"ethnicity": "French"},
            "@v": str(fr_lang.id) + "::" + str(fr_lang.revision_id),
        }
    ]

    # Commit clean dereferened fields
    # ---------------------------------------------------------------
    record.commit()
    assert record["languages"] == [{"id": str(en_lang.id)}]
    assert record["inner"]["inner_languages"][0]["id"] == str(fr_lang.id)

    # Commit to DB and refetch
    # ---------------------------------------------------------------
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record["languages"] == [{"id": str(en_lang.id)}]
    assert record["inner"]["inner_languages"][0]["id"] == str(fr_lang.id)
    _assert_iter_and_lang_from_record(record)

    # Clear field
    # ---------------------------------------------------------------
    record.relations.languages = None
    assert "languages" not in record
    assert record.relations.languages() is None

    record.relations.inner_languages = None
    assert "inner_languages" not in record
    assert record.relations.inner_languages() is None

    record.commit()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relation manager instance returned from the record."""


class RelationsMapping:
    """Helper class for managing relation fields."""

    def __init__(self, record, fields):
        """Initialize the relations mapping."""
        # Needed because we overwrite __setattr__
        cache = {}
        super().__setattr__("_record", record)
        super().__setattr__("_fields", fields)
        super().__setattr__("_cache", cache)
        for name, field in fields.items():
            field.inject_cache(cache, name)

    def __getattr__(self, name):
        """Get a relation field."""
        if name in self._fields:
            return self._fields[name].get_value(self._record)
        else:
            raise AttributeError

    def __setattr__(self, name, value):
        """Set a relation field."""
        if name in self._fields:
            field = self._fields[name]
            if value is None:
                field.clear_value(self._record)
            else:
                field.set_value(self._record, value)
        else:
            raise AttributeError

    def __delattr__(self, name):
        """Clear a relation field."""
        setattr(self, name, None)

    def __contains__(self, name):
        """Check if a field is in the mapping."""
        return name in self._fields

    def __iter__(self):
        """Iterate over the relations fields."""
        return iter(self._fields)

    def validate(self, fields=None):
        """Validates all relations in the record."""
        for name in fields or self:
            getattr(self, name).validate()

    def dereference(self, fields=None):
        """Dereferences relation fields."""
        for name in fields or self:
            getattr(self, name).dereference()

    def clean(self, fields=None):
        """Clean dereferenced relation fields."""
        for name in fields or self:
            getattr(self, name).clean()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2021 CERN.
# Copyright (C) 2021      TU Wien.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module for metadata storage."""


from functools import lru_cache

from invenio_base.utils import obj_or_import_string
from jsonref import JsonRef
from jsonresolver import JSONResolver
from jsonresolver.contrib.jsonref import json_loader_factory
from jsonresolver.contrib.jsonschema import ref_resolver_factory
from jsonschema import validate

from invenio_records.errors import RecordsRefResolverConfigError
from invenio_records.resolver import urljoin_with_custom_scheme

from . import config
from .validators import _create_validator


class _RecordsState(object):
    """State for record JSON resolver."""

    def __init__(self, app, entry_point_group=None):
        """Initialize state."""
        self.app = app
        self.resolver = JSONResolver(entry_point_group=entry_point_group)
        self.refresolver_cls = ref_resolver_factory(self.resolver)
        self.refresolver_store = None
        if self.app.config.get("RECORDS_REFRESOLVER_CLS"):
            self.refresolver_cls = obj_or_import_string(
                self.app.config.get("RECORDS_REFRESOLVER_CLS"),
            )
            self.refresolver_store = obj_or_import_string(
                self.app.config.get("RECORDS_REFRESOLVER_STORE")
            )

        self.loader_cls = json_loader_factory(self.resolver)

    def validate(self, data, schema, **kwargs):
        """Validate data using schema with ``JSONResolver``."""
        if not isinstance(schema, dict):
            schema = {"$ref": schema}
        refresolver_cls_kwargs = {}
        if self.refresolver_store:
            refresolver_cls_kwargs["store"] = self.refresolver_store
            refresolver_cls_kwargs["urljoin_cache"] = lru_cache(1024)(
                urljoin_with_custom_scheme
            )

        validator_cls = _create_validator(
            schema=schema,
            base_validator_cls=kwargs.pop("cls", None),
            custom_checks=self.app.config.get("RECORDS_VALIDATION_TYPES", {}),
        )

        resolver = self.refresolver_cls.from_schema(schema, **refresolver_cls_kwargs)

        return validate(data, schema, cls=validator_cls, resolver=resolver, **kwargs)

    def replace_refs(self, data):
        """Replace the JSON reference objects with ``JsonRef``."""
        return JsonRef.replace_refs(data, loader=self.loader_cls())


class InvenioRecords(object):
    """Invenio-Records extension."""

    def __init__(self, app=None, **kwargs):
        """Extension initialization."""
        if app:
            self._state = self.init_app(app, **kwargs)

    def init_app(self, app, entry_point_group="invenio_records.jsonresolver", **kwargs):
        """Flask application initialization.

        :param app: The Flask application.
        :param entry_point_group: The entrypoint for jsonresolver extensions.
            (Default: ``'invenio_records.jsonresolver'``)
        """
        self.init_config(app)
        state = _RecordsState(app, entry_point_group=entry_point_group)
        app.extensions["invenio-records"] = state
        return state

    def init_config(self, app):
        """Initialize configuration.

        :param app: The Flask application.
        """
        app.config.setdefault("SQLALCHEMY_TRACK_MODIFICATIONS", True)
        for k in dir(config):
            if k.startswith("RECORDS_"):
                app.config.setdefault(k, getattr(config, k))

        if app.config.get("RECORDS_REFRESOLVER_CLS") and not app.config.get(
            "RECORDS_REFRESOLVER_STORE"
        ):
            raise RecordsRefResolverConfigError(
                "RECORDS_REFRESOLVER_CLS config requires "
                "RECORDS_REFRESOLVER_STORE to be set."
            )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test of dictionary utilities."""

from copy import deepcopy

import pytest

from invenio_records.dictutils import clear_none, dict_lookup, dict_merge


def test_clear_none():
    """Test clearning of the dictionary."""
    d = {
        "a": None,
        "b": {"c": None},
        "d": ["1", None, []],
        "e": [{"a": None, "b": []}],
    }

    clear_none(d)
    # Modifications are done in place, so gotta test after the function call.
    assert d == {"d": ["1"]}

    d = {
        "a": None,
        "b": [
            {"a": "1", "b": None},
        ],
    }
    clear_none(d)
    # Modifications are done in place, so gotta test after the function call.
    assert d == {"b": [{"a": "1"}]}


def test_dict_lookup():
    """Test lookup by a key."""
    d = {
        "a": 1,
        "b": {"c": None},
        "d": ["1", "2"],
    }
    assert dict_lookup(d, "a") == d["a"]
    assert dict_lookup(d, "b") == d["b"]
    assert dict_lookup(d, "b.c") == d["b"]["c"]
    assert dict_lookup(d, "d") == d["d"]
    assert dict_lookup(d, "d.0") == d["d"][0]
    assert dict_lookup(d, "d.1") == d["d"][1]
    assert dict_lookup(d, "d.-1") == d["d"][-1]

    assert pytest.raises(KeyError, dict_lookup, d, "x")
    assert pytest.raises(KeyError, dict_lookup, d, "a.x")
    assert pytest.raises(KeyError, dict_lookup, d, "b.x")
    assert pytest.raises(KeyError, dict_lookup, d, "b.c.0")
    assert pytest.raises(KeyError, dict_lookup, d, "d.3")


def test_dict_merge():
    """Test dict merge."""
    dest = dict()
    source = dict(foo="bar")
    dict_merge(dest, source)

    assert dest == dict(foo="bar")

    dest = {"foo1": "bar1", "metadata": {"field1": 3}}
    source = {"foo2": "bar2", "metadata": {"field2": "test"}}
    dict_merge(dest, source)

    assert dest == {
        "foo1": "bar1",
        "foo2": "bar2",
        "metadata": {"field1": 3, "field2": "test"},
    }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default values for records configuration."""

RECORDS_VALIDATION_TYPES = {}
"""Pass additional types when validating a record against a schema.
For more details, see:
`<https://python-jsonschema.readthedocs.io/en/latest/validate/#validating-types>`_.
"""

RECORDS_REFRESOLVER_CLS = None
"""Custom JSONSchemas ref resolver class.

Note that when using a custom ref resolver class you should also set
``RECORDS_REFRESOLVER_STORE`` to point to a JSONSchema ref resolver store.
"""

RECORDS_REFRESOLVER_STORE = None
"""JSONSchemas ref resolver store.

Used together with ``RECORDS_REFRESOLVER_CLS`` to provide a specific
ref resolver store.
"""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Extensions allow integration of features into a record class.

For instance, the system fields feature is built as an extension.
"""


class ExtensionMixin:
    """Defines the methods needed by an extension."""

    def pre_init(self, record, data, model=None, **kwargs):
        """Called when a new record instance is initialized.

        Called when a new record is instantiated (i.e. during all
        ``Record({...})``). This means it's also called when e.g. a record
        is created via ``Record.create()``.

        :param data: The dict passed to the record's constructor.
        :param model: The model class used for initialization.
        """

    def post_init(self, record, data, model=None, **kwargs):
        """Called after a record is initialized."""

    def pre_dump(self, record, data, dumper=None):
        """Called before a record is dumped."""

    def post_dump(self, record, data, dumper=None):
        """Called after a record is dumped."""

    def pre_load(self, data, loader=None):
        """Called before a record is loaded."""

    def post_load(self, record, data, loader=None):
        """Called after a record is loaded."""

    def pre_create(self, record):
        """Called before a record is created."""

    def post_create(self, record):
        """Called after a record is created."""

    def pre_commit(self, record):
        """Called before a record is committed."""

    def post_commit(self, record):
        """Called after a record is committed."""

    def pre_delete(self, record, force=False):
        """Called before a record is deleted."""

    def post_delete(self, record, force=False):
        """Called after a record is deleted."""

    def pre_revert(self, record, revision):
        """Called before a record is reverted."""

    def post_revert(self, new_record, revision):
        """Called after a record is reverted."""


class RecordExtension(ExtensionMixin):
    """Base class for a record extensions."""


class RecordMeta(type):
    """Metaclass responsible for initializing the extension registry."""

    def __new__(mcs, name, bases, attrs):
        """Create a new record class."""
        # Initialise an "_extensions" attribute on each class, to ensure each
        # class have a separate extensions registry.
        if "_extensions" not in attrs:
            attrs["_extensions"] = []

        # Construct the class.
        return super().__new__(mcs, name, bases, attrs)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records."""


import json
import os
import uuid

import pytest
from flask import Flask
from invenio_db.utils import drop_alembic_version_table
from jsonschema.exceptions import ValidationError
from sqlalchemy.orm.exc import NoResultFound

from invenio_records import InvenioRecords, Record
from invenio_records.errors import MissingModelError


def test_version():
    """Test version import."""
    from invenio_records import __version__

    assert __version__


def test_init():
    """Test extension initialization."""
    app = Flask("testapp")
    ext = InvenioRecords(app)
    assert "invenio-records" in app.extensions

    app = Flask("testapp")
    ext = InvenioRecords()
    assert "invenio-records" not in app.extensions
    ext.init_app(app)
    assert "invenio-records" in app.extensions


def test_db(testapp, db):
    """Test database backend."""
    with testapp.app_context():
        assert "records_metadata" in db.metadata.tables
        assert "records_metadata_version" in db.metadata.tables
        assert "transaction" in db.metadata.tables

    schema = {
        "type": "object",
        "properties": {
            "title": {"type": "string"},
            "field": {"type": "boolean"},
            "hello": {"type": "array"},
        },
        "required": ["title"],
    }
    data = {"title": "Test", "$schema": schema}
    from invenio_records.models import RecordMetadata as RM

    # Create a record
    with testapp.app_context():
        assert RM.query.count() == 0

        record_uuid = Record.create(data).id
        db.session.commit()

        assert RM.query.count() == 1
        db.session.commit()

    # Retrieve created record
    with testapp.app_context():
        record = Record.get_record(record_uuid)
        assert record.dumps() == data
        with pytest.raises(NoResultFound):
            Record.get_record(uuid.uuid4())
        record["field"] = True
        record = record.patch([{"op": "add", "path": "/hello", "value": ["world"]}])
        assert record["hello"] == ["world"]
        record.commit()
        db.session.commit()

    with testapp.app_context():
        record2 = Record.get_record(record_uuid)
        assert record2.model.version_id == 2
        assert record2["field"]
        assert record2["hello"] == ["world"]
        db.session.commit()

    # Cannot commit record without model (i.e. Record.create_record)
    with testapp.app_context():
        record3 = Record({"title": "Not possible"})
        with pytest.raises(MissingModelError):
            record3.commit()

    # Check invalid schema values
    with testapp.app_context():
        data = {
            "$schema": "http://json-schema.org/learn/examples/"
            "geographical-location.schema.json",
            "latitude": 42,
            "longitude": 42,
        }

        record_with_schema = Record.create(data).commit()
        db.session.commit()

        record_with_schema["latitude"] = "invalid"
        with pytest.raises(ValidationError):
            record_with_schema.commit()

    # Allow types overriding on schema validation
    with testapp.app_context():
        data = {"title": "Test", "hello": tuple(["foo", "bar"]), "$schema": schema}
        testapp.config["RECORDS_VALIDATION_TYPES"] = {}
        with pytest.raises(ValidationError):
            Record.create(data).commit()

        testapp.config["RECORDS_VALIDATION_TYPES"] = {"array": (list, tuple)}
        record_uuid = Record.create(data).commit()
        db.session.commit()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create records tables."""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op

# revision identifiers, used by Alembic.
revision = "862037093962"
down_revision = "1095cdf9f350"
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.create_table(
        "records_metadata",
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("id", sqlalchemy_utils.types.uuid.UUIDType(), nullable=False),
        sa.Column(
            "json",
            sqlalchemy_utils.JSONType().with_variant(
                sa.dialects.postgresql.JSON(none_as_null=True),
                "postgresql",
            ),
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "records_metadata_version",
        sa.Column("created", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column("updated", sa.DateTime(), autoincrement=False, nullable=True),
        sa.Column(
            "id",
            sqlalchemy_utils.types.uuid.UUIDType(),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column(
            "json",
            sqlalchemy_utils.JSONType().with_variant(
                sa.dialects.postgresql.JSON(none_as_null=True),
                "postgresql",
            ),
            autoincrement=False,
            nullable=True,
        ),
        sa.Column("version_id", sa.Integer(), autoincrement=False, nullable=True),
        sa.Column(
            "transaction_id", sa.BigInteger(), autoincrement=False, nullable=False
        ),
        sa.Column("end_transaction_id", sa.BigInteger(), nullable=True),
        sa.Column("operation_type", sa.SmallInteger(), nullable=False),
        sa.PrimaryKeyConstraint("id", "transaction_id"),
    )
    op.create_index(
        op.f("ix_records_metadata_version_end_transaction_id"),
        "records_metadata_version",
        ["end_transaction_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_records_metadata_version_operation_type"),
        "records_metadata_version",
        ["operation_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_records_metadata_version_transaction_id"),
        "records_metadata_version",
        ["transaction_id"],
        unique=False,
    )


def downgrade():
    """Downgrade database."""
    op.drop_index(
        op.f("ix_records_metadata_version_transaction_id"),
        table_name="records_metadata_version",
    )
    op.drop_index(
        op.f("ix_records_metadata_version_operation_type"),
        table_name="records_metadata_version",
    )
    op.drop_index(
        op.f("ix_records_metadata_version_end_transaction_id"),
        table_name="records_metadata_version",
    )
    op.drop_table("records_metadata_version")
    op.drop_table("records_metadata")



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""The core of the system fields implementation."""

import inspect
import warnings

from ..dictutils import dict_lookup, parse_lookup_key
from ..extensions import ExtensionMixin, RecordExtension, RecordMeta


def _get_fields(attrs, field_class):
    """Get system fields from a class' attributes dict.

    :param attrs: Dict of class' attributes.
    :param field_class: Base class for all system fields.
    """
    fields = {}
    for name, val in attrs.items():
        if isinstance(val, field_class):
            fields[name] = val
    return fields


def _get_inherited_fields(class_, field_class):
    """Get system fields from all base classes respecting the MRO.

    :param class_: The class that has just been constructed.
    :param field_class: Base class for all system fields.
    """
    # The variable mro holds a list of all super classes of "class_", in
    # correct MRO (mehtod resolution order). This ensures that we respect
    # Python inheritance rules when the same field is defined in both
    # the class and one or more super classes.
    mro = inspect.getmro(class_)

    fields = {}

    # Reverse order through MRO so that a field is taken from the most
    # specific class.
    for base in reversed(mro):
        fields.update(_get_fields(base.__dict__, field_class))
    return fields


class SystemFieldContext:
    """Base class for a system field context.

    A system field context is created once you access a field's attribute on
    a class. As the system field may be defined on a super class, this context
    allows us to know from which class the field was accessed.

    Normally you should subclass this class, and implement methods the methods
    on it that requires you to know the record class.
    """

    def __init__(self, field, record_cls):
        """Initialise the field context."""
        self._field = field
        self._record_cls = record_cls

    @property
    def field(self):
        """Access the field to prevent it from being overwritten."""
        return self._field

    @property
    def record_cls(self):
        """Record class to prevent it from being overwritten."""
        return self._record_cls


class SystemField(ExtensionMixin):
    """Base class for all system fields.

    A system field is a Python data descriptor set on a record class that can
    also hook into a record via the extensions API (e.g on record creation,
    dumping etc).


    See :py:class:`~invenio_records.extensions.ExtensionMixin` for the full
    interface of methods that a field can override to hook into the record
    API.
    """

    def __init__(self, key=None):
        """Initialise the field."""
        self._key = key
        # The attribute is set by __set_name__ which is called by the metaclass
        # during construction.
        self._attr_name = None

    @property
    def attr_name(self):
        """Property to access the assigned class attribute name.

        :returns: ``None`` if field is not assigned, otherwise the class
                  attribute name.
        """
        return self._attr_name

    @property
    def key(self):
        """Property to access the dict key name.

        Uses the attribute name if the key is not defined.
        """
        return self._key or self._attr_name

    #
    # Data descriptor definition
    #
    def __get__(self, record, owner=None):
        r"""Accessing the object attribute.

        A subclass that overwrites this method, should handle two cases:

        1. Class access - If ``instance`` is None, the field is accessed
           through the class (e.g. Record.myfield). In this case a field or
           context should be returned. The purpose of the field context, is
           to allow a field to know from which class it was accessed (as the
           field may be created on a super class).
        2. Instance access -  If ``instance`` is not None, the field is
           accessed through an instance of the class (e.g. record``.myfield``).

        A simple example is provided below:

        .. code-block:: python

            def __get__(self, record, owner=None):
                if record is None:
                    return self
                    return SystemFieldContext(self, owner)
                if 'mykey' in record:
                    return record['mykey']
                return None

        :param record: The instance through which the field is being accessed
                       or ``None`` if the field is accessed through the
                       class.
        :param owner: The class which owns the field.
        """
        # Class access
        # - by default a system field accessed through a class (e.g.
        #   Record.myattr will return a field or field context).
        if record is None:
            return self
        # Instance access
        # - by default a system field accessed through an object instance (e.g.
        #   record.myattr will raise an Attribute error)
        raise AttributeError

    def __set__(self, record, value):
        """Setting the attribute (instance access only).

        This method only handles set operations from an instance (e.g.
        ``record.myfield = val``). This is opposite to ``__get__()`` which
        needs to handle both class and instance access.
        """
        raise AttributeError()

    def __set_name__(self, owner, name):
        """Inject the class attribute name into the field.

        This ensures that a field has access to the attribute name used on the
        class. In the following example, the attribute name ``schema`` would be
        set in the ``ConstantField`` object instance.

        .. code-block:: python

            class MyRecord(Record, SystemFieldsMixin):
                schema = ConstantField(...)
        """
        self._attr_name = name

    def post_init(self, record, data, model=None, field_data=None):
        """Core implementation of post_init to support argument loading."""
        if field_data is not None:
            self.__set__(record, field_data)

    def get_dictkey(self, instance):
        """Helper to use a lookup key to get a nested object.

        Assume the key have been set in ``self.key``
        """
        try:
            return dict_lookup(instance, self.key)
        except KeyError:
            return None

    def set_dictkey(self, instance, value, create_if_missing=False):
        """Helper to set value using a lookup key on a nested object."""
        keys = parse_lookup_key(self.key)
        try:
            parent = dict_lookup(instance, keys, parent=True)
        except KeyError as e:
            if not create_if_missing:
                raise
            parent = instance
            for k in keys[:-1]:
                if k not in parent:
                    parent[k] = {}
                else:
                    if not isinstance(parent[k], dict):
                        raise KeyError(
                            "Expected a dict at subkey '{}'. "
                            "Found '{}'.".format(k, parent[k].__class__.__name__)
                        )
                parent = parent[k]

        if not isinstance(parent, dict):
            raise KeyError(
                "Expected a dict at subkey '{}'. Found '{}'.".format(
                    keys[-2], parent.__class__.__name__
                )
            )

        parent[keys[-1]] = value

    def _set_cache(self, instance, obj):
        """Set an object on the instance's cache."""
        if not hasattr(instance, "_obj_cache"):
            instance._obj_cache = {}
        instance._obj_cache[self.attr_name] = obj

    def _get_cache(self, instance):
        """Get the object from the instance's cache."""
        return getattr(instance, "_obj_cache", {}).get(self.attr_name)


class SystemFieldsExt(RecordExtension):
    """Record extension for system fields.

    This extension is responsible for iterating over all declared system fields
    on a class for each extension point.
    """

    def __init__(self, declared_fields):
        """Save the declared fields on the extension."""
        self.declared_fields = declared_fields

    def _run(self, method, *args, **kwargs):
        for field in self.declared_fields.values():
            getattr(field, method)(*args, **kwargs)

    def pre_init(self, *args, **kwargs):
        """Called before a record is dumped."""
        self._run("pre_init", *args, **kwargs)

    def post_init(self, record, data, model=None, **kwargs):
        """Called when a new record instance is initialized."""
        # Special treatment for post_init (also has special implementation
        # in SystemField)
        for attr_name, field in self.declared_fields.items():
            field_data = kwargs.get(field.attr_name)
            field.post_init(record, data, model=model, field_data=field_data)

    def pre_dump(self, record, data, dumper=None):
        """Called before a record is dumped."""
        for field in self.declared_fields.values():
            pre_dump_params = inspect.signature(field.pre_dump).parameters
            if "data" in pre_dump_params:
                field.pre_dump(record, data, dumper=dumper)
            else:
                # TODO: Remove in v1.6.0 or later
                warnings.warn(
                    "The pre_dump hook must take a positional argument data.",
                    DeprecationWarning,
                )
                field.pre_dump(record, dumper=dumper)

    def post_dump(self, record, data, dumper=None):
        """Called after a record is dumped."""
        for field in self.declared_fields.values():
            post_dump_params = inspect.signature(field.post_dump).parameters
            if "data" in post_dump_params:
                field.post_dump(record, data, dumper=dumper)
            else:
                # TODO: Remove in v1.6.0 or later
                warnings.warn(
                    "The post_dump hook must take a positional argument data.",
                    DeprecationWarning,
                )
                field.post_dump(record, dumper=dumper)

    def pre_load(self, data, loader=None):
        """Called before a record is loaded."""
        for field in self.declared_fields.values():
            pre_load_params = inspect.signature(field.pre_load).parameters
            if "data" in pre_load_params:
                field.pre_load(data, loader=loader)
            else:
                # TODO: Remove in v1.6.0 or later
                warnings.warn(
                    "The pre_load hook must take a positional argument data.",
                    DeprecationWarning,
                )
                field.pre_load(loader=loader)

    def post_load(self, record, data, loader=None):
        """Called after a record is loaded."""
        for field in self.declared_fields.values():
            post_load_params = inspect.signature(field.post_load).parameters
            if "data" in post_load_params:
                field.post_load(record, data, loader=loader)
            else:
                # TODO: Remove in v1.6.0 or later
                warnings.warn(
                    "The post_load hook must take a positional argument data.",
                    DeprecationWarning,
                )
                field.post_load(record, loader=loader)

    def pre_create(self, *args, **kwargs):
        """Called after a record is created."""
        self._run("pre_create", *args, **kwargs)

    def post_create(self, *args, **kwargs):
        """Called after a record is created."""
        self._run("post_create", *args, **kwargs)

    def pre_commit(self, *args, **kwargs):
        """Called before a record is committed."""
        self._run("pre_commit", *args, **kwargs)

    def post_commit(self, *args, **kwargs):
        """Called after a record is commited."""
        self._run("post_commit", *args, **kwargs)

    def pre_delete(self, *args, **kwargs):
        """Called before a record is deleted."""
        self._run("pre_delete", *args, **kwargs)

    def post_delete(self, *args, **kwargs):
        """Called after a record is deleted."""
        self._run("post_delete", *args, **kwargs)

    def pre_revert(self, *args, **kwargs):
        """Called before a record is reverted."""
        self._run("pre_revert", *args, **kwargs)

    def post_revert(self, *args, **kwargs):
        """Called after a record is reverted."""
        self._run("post_revert", *args, **kwargs)


class SystemFieldsMeta(RecordMeta):
    """Metaclass for a record class that integrates system fields."""

    def __new__(mcs, name, bases, attrs):
        """Create a new record class."""
        # Construct the class (and initialise the extension registry).
        class_ = super().__new__(mcs, name, bases, attrs)

        # Get system fields and ensure inheritance is respected.
        declared_fields = _get_inherited_fields(class_, SystemField)
        declared_fields.update(_get_fields(attrs, SystemField))

        # Register the system fields extension on the record class.
        class_._extensions.append(SystemFieldsExt(declared_fields))

        return class_


class SystemFieldsMixin(metaclass=SystemFieldsMeta):
    """Mixin class for records that add system fields capabilities.

    This class is primarily syntax sugar for being able to do::

      class MyRecord(Record, SystemsFieldsMixin):
          pass

    instead of::

      class MyRecord(Record, metaclass=SystemFieldsMeta):
          pass

    There are subtle differences though between the two above methods. Mainly
    which classes will execute the ``__new__()`` method on the metaclass.
    """



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Constant system field."""

from ..dictutils import dict_lookup
from .base import SystemField


class ConstantField(SystemField):
    """Constant fields add a constant value to a key in the record."""

    def __init__(self, key=None, value=""):
        """Initialize the field.

        :param key: The key to set in the dictionary (dot notation supported
                    for nested lookup).
        :param value: The value to set for the key.
        """
        self.value = value
        super().__init__(key=key)

    def pre_init(self, record, data, model=None, **kwargs):
        """Sets the key in the record during record instantiation."""
        if data is None:
            # A deleted record.
            return
        try:
            dict_lookup(data, self.key)
        except KeyError:
            # Key is not present, so add it.
            data[self.key] = self.value

    def __get__(self, record, owner=None):
        """Accessing the attribute."""
        # Class access
        if record is None:
            return self
        # Instance access
        try:
            return dict_lookup(record, self.key)
        except KeyError:
            return None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests for relations system field."""

from collections.abc import Iterable

import pytest

from invenio_records.api import Record
from invenio_records.systemfields import PKRelation, RelationsField, SystemFieldsMixin
from invenio_records.systemfields.relations import (
    InvalidRelationValue,
    PKListRelation,
    PKNestedListRelation,
    RelationsMapping,
)
from invenio_records.systemfields.relations.errors import InvalidCheckValue


def test_relations_field_pk_relation(testapp, db, languages):
    """RelationsField tests for PKRelation."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(key="language", keys=["iso"], record_cls=Language),
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.language, PKRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.language)
    assert record.relations.language() is None

    # Initialize from dictionary data
    record = Record1({"language": {"id": str(en_lang.id)}})
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    # Initialize from field data
    record = Record1({}, relations={"language": str(en_lang.id)})
    assert record["language"]["id"] == str(en_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    # Set via record data
    record = Record1.create({})
    record["language"] = {"id": str(en_lang.id)}
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang
    record.commit()  # validates
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    record["language"] = {"id": "invalid"}
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.language = str(en_lang.id)
    assert record["language"]["id"] == str(en_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == en_lang

    # Update existing value
    record.relations.language = str(fr_lang.id)
    assert record["language"]["id"] == str(fr_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.language = "invalid"
    # Check that old value is still there
    assert record["language"]["id"] == str(fr_lang.id)
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    # Dereference relation
    record.relations.language.dereference()
    assert record["language"] == {
        "id": str(fr_lang.id),
        "iso": "fr",  # only stores the "iso" field
        "@v": str(fr_lang.id) + "::" + str(fr_lang.revision_id),
    }

    # Commit clean dereferened fields
    record.commit()
    assert record["language"] == {"id": str(fr_lang.id)}

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record["language"] == {"id": str(fr_lang.id)}
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    # Clear field
    record.relations.language = None
    assert "language" not in record
    assert record.relations.language() is None
    record.commit()


def test_relations_field_pk_list_relation(testapp, db, languages):
    """RelationsField tests for PKListRelation."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key="languages",
                keys=["iso", "information.ethnicity"],
                record_cls=Language,
            ),
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.languages, PKListRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.languages)
    assert record.relations.languages() is None

    # Initialize from dictionary data
    record = Record1({"languages": [{"id": str(en_lang.id)}]})
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Initialize from field data
    record = Record1({}, relations={"languages": [str(en_lang.id)]})
    assert record["languages"] == [{"id": str(en_lang.id)}]
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Set via record dictionary data
    record = Record1.create({})
    record["languages"] = [{"id": str(en_lang.id)}]
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    record.commit()  # validates
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    record["languages"] = {"id": "invalid"}
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.languages = [str(en_lang.id)]
    assert record["languages"][0]["id"] == str(en_lang.id)
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang

    # Update existing value
    record.relations.languages = [str(fr_lang.id)]
    assert record["languages"][0]["id"] == str(fr_lang.id)
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    # Set invalid value
    for v in ("invalid", ["invalid"]):
        with pytest.raises(InvalidRelationValue):
            record.relations.languages = v
        # Check that old value is still there
        assert record["languages"][0]["id"] == str(fr_lang.id)
        res_iter = record.relations.languages()
        assert isinstance(res_iter, Iterable)
        res = next(res_iter)
        assert res == fr_lang

    # Dereference relation
    record.relations.languages.dereference()
    assert record["languages"] == [
        {
            "id": str(fr_lang.id),
            "iso": "fr",  # only stores the "iso" field
            "information": {"ethnicity": "French"},
            "@v": str(fr_lang.id) + "::" + str(fr_lang.revision_id),
        }
    ]

    # Commit clean dereferened fields
    record.commit()
    assert record["languages"] == [{"id": str(fr_lang.id)}]

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record["languages"] == [{"id": str(fr_lang.id)}]
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    # Clear field
    record.relations.languages = None
    assert "languages" not in record
    assert record.relations.languages() is None
    record.commit()


def test_relations_field_pk_list_relation_of_objects(testapp, db, languages):
    """RelationsField tests for PKListRelation with a list of objects."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            array_of_objects=PKListRelation(
                key="array_of_objects",
                keys=["iso"],
                record_cls=Language,
                relation_field="language",
            ),
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.array_of_objects, PKListRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.array_of_objects)
    assert record.relations.array_of_objects() is None

    # Initialize from dictionary data
    record = Record1(
        {
            "array_of_objects": [
                {
                    "field1": "no language",
                    "field2": "random string",
                    "field3": {"nestedField": "nested string"},
                },
                {"language": {"id": str(en_lang.id)}, "field1": "string"},
                {"language": {"id": str(fr_lang.id)}, "field2": "string2"},
                {"language": {"id": str(en_lang.id)}, "field1": "string"},
            ]
        }
    )

    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang
    res = next(res_iter)
    assert res == en_lang

    # Initialize from field data
    record = Record1(
        {"metadata": {"title": "random title"}},
        relations={
            "array_of_objects": [
                {
                    "field1": "no language",
                    "field2": "random string",
                    "field3": {"nestedField": "nested string"},
                },
                {"language": str(en_lang.id), "field1": "string"},
                {"language": str(fr_lang.id), "field2": "string2"},
                {"language": str(en_lang.id), "field1": "string"},
            ]
        },
    )
    with pytest.raises(KeyError):
        record["array_of_objects"][0]["language"]["id"]
    assert record["array_of_objects"][1]["language"]["id"] == str(en_lang.id)
    assert record["array_of_objects"][2]["language"]["id"] == str(fr_lang.id)
    assert record["array_of_objects"][3]["language"]["id"] == str(en_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang
    res = next(res_iter)
    assert res == en_lang

    # Set via record dictionary data
    record = Record1.create({})
    record["array_of_objects"] = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
        },
        {"language": {"id": str(en_lang.id)}, "field1": "string"},
        {"language": {"id": str(fr_lang.id)}, "field2": "string2"},
        {"language": {"id": str(en_lang.id)}, "field1": "string"},
    ]
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang

    record.commit()  # validates
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang

    record["array_of_objects"] = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
            "language": {"id": "invalid"},
        },
        {"field1": "testString"},
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.array_of_objects = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
        },
        {"language": str(en_lang.id), "field1": "string"},
        {"language": str(fr_lang.id), "field2": "string2"},
        {"language": str(en_lang.id), "field1": "string"},
    ]

    assert record["array_of_objects"][1]["language"]["id"] == str(en_lang.id)
    assert record["array_of_objects"][2]["language"]["id"] == str(fr_lang.id)
    assert record["array_of_objects"][3]["language"]["id"] == str(en_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == fr_lang
    res = next(res_iter)
    assert res == en_lang

    # Update existing value
    record.relations.array_of_objects = [
        {
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
            "language": str(en_lang.id),
        },
        {"language": str(es_lang.id), "field1": "string"},
        {"language": str(en_lang.id), "field1": "string"},
    ]
    assert record["array_of_objects"][0]["language"]["id"] == str(en_lang.id)
    assert record["array_of_objects"][1]["language"]["id"] == str(es_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == es_lang

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.array_of_objects = "invalid"
    # Check that old value is still there
    assert record["array_of_objects"][0]["language"]["id"] == str(en_lang.id)
    assert record["array_of_objects"][1]["language"]["id"] == str(es_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == es_lang

    # Dereference relation
    record.relations.array_of_objects.dereference()
    assert record["array_of_objects"] == [
        {
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
            "language": {
                "id": str(en_lang.id),
                "iso": "en",
                "@v": str(en_lang.id) + "::" + str(en_lang.revision_id),
            },
        },
        {
            "language": {
                "id": str(es_lang.id),
                "iso": "es",
                "@v": str(es_lang.id) + "::" + str(es_lang.revision_id),
            },
            "field1": "string",
        },
        {
            "language": {
                "id": str(en_lang.id),
                "iso": "en",
                "@v": str(en_lang.id) + "::" + str(en_lang.revision_id),
            },
            "field1": "string",
        },
    ]

    # Commit clean dereferenced fields
    record.commit()
    assert record["array_of_objects"][0]["language"]["id"] == str(en_lang.id)
    assert record["array_of_objects"][1]["language"]["id"] == str(es_lang.id)
    assert record["array_of_objects"][2]["language"]["id"] == str(en_lang.id)

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record["array_of_objects"][0]["language"]["id"] == str(en_lang.id)
    assert record["array_of_objects"][1]["language"]["id"] == str(es_lang.id)
    assert record["array_of_objects"][2]["language"]["id"] == str(en_lang.id)
    res_iter = record.relations.array_of_objects()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == en_lang
    res = next(res_iter)
    assert res == es_lang
    res = next(res_iter)
    assert res == en_lang

    # Clear field
    record.relations.array_of_objects = None
    assert "array_of_objects" not in record
    assert record.relations.array_of_objects() is None
    record.commit()


def test_nested_relations_field(testapp, db, languages):
    """Tests deeply nested relations."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]
    it_lang = languages["it"]

    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            deep_single=PKRelation(key="metadata.deep.language", record_cls=Language),
            deep_list=PKListRelation(
                key="metadata.deep.languages", record_cls=Language
            ),
            deep_sibling=PKRelation(
                key="metadata.deep.dark.language", record_cls=Language
            ),
        )

    # Set all fields
    record = Record1.create({})
    record.relations.deep_single = en_lang
    record.relations.deep_list = [en_lang, fr_lang, es_lang]
    record.relations.deep_sibling = it_lang

    assert record == {
        "metadata": {
            "deep": {
                "language": {"id": str(en_lang.id)},
                "languages": [
                    {"id": str(en_lang.id)},
                    {"id": str(fr_lang.id)},
                    {"id": str(es_lang.id)},
                ],
                "dark": {
                    "language": {"id": str(it_lang.id)},
                },
            }
        }
    }
    record.commit()


def test_relations_field_pk_nested_list_of_objects_w_related_field(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with a list of objects."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]

    class Record1(Record, SystemFieldsMixin):
        # end result should be
        # {
        #   nested_array: [
        #       {languages: [en, fr]},
        #       {languages: [en, es]}
        #   ]
        # }
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key="nested_array_of_objects",
                keys=["iso"],
                record_cls=Language,
                relation_field="languages",
            ),
        )

    def _test_iterators(record):
        res_iter = record.relations.nested_array_of_objects()
        assert isinstance(res_iter, Iterable)
        # first inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == en_lang
        res = next(res_inner_iter)
        assert res == es_lang
        with pytest.raises(StopIteration):  # finished first iterator
            res = next(res_inner_iter)
        # second inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == fr_lang
        res = next(res_inner_iter)
        assert res == en_lang
        with pytest.raises(StopIteration):  # finished second iterator
            res = next(res_inner_iter)

    def _test_inner_values(record):
        with pytest.raises(KeyError):
            record["nested_array_of_objects"][0]["languages"][0]["id"]
        assert record["nested_array_of_objects"][1]["languages"][0]["id"] == (
            str(en_lang.id)
        )
        assert record["nested_array_of_objects"][1]["languages"][1]["id"] == (
            str(es_lang.id)
        )
        assert record["nested_array_of_objects"][2]["languages"][0]["id"] == (
            str(fr_lang.id)
        )
        assert record["nested_array_of_objects"][2]["languages"][1]["id"] == (
            str(en_lang.id)
        )

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.nested_array_of_objects, PKNestedListRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.nested_array_of_objects)
    assert record.relations.nested_array_of_objects() is None

    # Initialize from dictionary data
    record = Record1(
        {
            "nested_array_of_objects": [
                {
                    "field1": "no language",
                    "field2": "random string",
                    "field3": {"nestedField": "nested string"},
                },
                {
                    "languages": [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
                    "field1": "string",
                },
                {
                    "languages": [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
                    "field2": "string2",
                },
            ]
        }
    )

    _test_iterators(record)

    # Initialize from field data
    record = Record1(
        {"metadata": {"title": "random title"}},
        relations={
            "nested_array_of_objects": [
                {
                    "field1": "no language",
                    "field2": "random string",
                    "field3": {"nestedField": "nested string"},
                },
                {"languages": [str(en_lang.id), str(es_lang.id)], "field1": "string"},
                {"languages": [str(fr_lang.id), str(en_lang.id)], "field2": "string2"},
            ]
        },
    )
    _test_inner_values(record)
    _test_iterators(record)

    # Set via record dictionary data
    record = Record1.create({})
    record["nested_array_of_objects"] = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
        },
        {
            "languages": [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
            "field1": "string",
        },
        {
            "languages": [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
            "field2": "string2",
        },
    ]

    _test_iterators(record)

    record.commit()  # validates

    _test_iterators(record)

    # fails with non-list types
    record["nested_array_of_objects"] = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
            "languages": {"id": str(es_lang.id)},
        },
        {"field1": "testString"},
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # fails with invalid values
    record["nested_array_of_objects"] = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
            "languages": [{"id": "invalid"}],
        },
        {"field1": "testString"},
    ]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.nested_array_of_objects = [
        {
            "field1": "no language",
            "field2": "random string",
            "field3": {"nestedField": "nested string"},
        },
        {"languages": [str(en_lang.id), str(es_lang.id)], "field1": "string"},
        {"languages": [str(fr_lang.id), str(en_lang.id)], "field2": "string2"},
    ]
    _test_inner_values(record)
    _test_iterators(record)

    # Update existing value
    record.relations.nested_array_of_objects = [
        {"languages": [str(fr_lang.id), str(en_lang.id)], "field2": "string2"}
    ]
    assert record["nested_array_of_objects"][0]["languages"][0]["id"] == (
        str(fr_lang.id)
    )
    assert record["nested_array_of_objects"][0]["languages"][1]["id"] == (
        str(en_lang.id)
    )
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.nested_array_of_objects = "invalid"
    # Check that old value is still there
    assert record["nested_array_of_objects"][0]["languages"][0]["id"] == (
        str(fr_lang.id)
    )
    assert record["nested_array_of_objects"][0]["languages"][1]["id"] == (
        str(en_lang.id)
    )
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Dereference relation
    record.relations.nested_array_of_objects.dereference()
    assert record["nested_array_of_objects"] == [
        {
            "languages": [  # last updated values are fr + en
                {
                    "id": str(fr_lang.id),
                    "iso": "fr",
                    "@v": str(fr_lang.id) + "::" + str(fr_lang.revision_id),
                },
                {
                    "id": str(en_lang.id),
                    "iso": "en",
                    "@v": str(en_lang.id) + "::" + str(en_lang.revision_id),
                },
            ],
            "field2": "string2",
        },
    ]

    # Commit clean dereferenced fields
    record.commit()
    assert record["nested_array_of_objects"][0]["languages"][0]["id"] == (
        str(fr_lang.id)
    )
    assert record["nested_array_of_objects"][0]["languages"][1]["id"] == (
        str(en_lang.id)
    )

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record["nested_array_of_objects"][0]["languages"][0]["id"] == (
        str(fr_lang.id)
    )
    assert record["nested_array_of_objects"][0]["languages"][1]["id"] == (
        str(en_lang.id)
    )
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Clear field
    record.relations.nested_array_of_objects = None
    assert "nested_array_of_objects" not in record
    assert record.relations.nested_array_of_objects() is None
    record.commit()


def test_relations_field_pk_nested_list_of_objects_wo_related_field(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with a list of objects."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]

    class Record1(Record, SystemFieldsMixin):
        # end result should be
        # {
        #   nested_languages: [
        #       [en, fr],
        #       [es, fr],
        #   ]
        # }
        relations = RelationsField(
            nested_languages=PKNestedListRelation(
                key="nested_languages",
                keys=["iso"],
                record_cls=Language,
            ),
        )

    def _test_iterators(record):
        res_iter = record.relations.nested_languages()
        assert isinstance(res_iter, Iterable)
        # first inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == en_lang
        res = next(res_inner_iter)
        assert res == es_lang
        with pytest.raises(StopIteration):  # finished first iterator
            res = next(res_inner_iter)
        # second inner list
        res_inner_iter = next(res_iter)
        assert isinstance(res_inner_iter, Iterable)
        res = next(res_inner_iter)
        assert res == fr_lang
        res = next(res_inner_iter)
        assert res == en_lang
        with pytest.raises(StopIteration):  # finished second iterator
            res = next(res_inner_iter)

    def _test_inner_values(record):
        assert record["nested_languages"][0][0]["id"] == (str(en_lang.id))
        assert record["nested_languages"][0][1]["id"] == (str(es_lang.id))
        assert record["nested_languages"][1][0]["id"] == (str(fr_lang.id))
        assert record["nested_languages"][1][1]["id"] == (str(en_lang.id))

    # Class-field check
    assert isinstance(Record1.relations, RelationsField)
    assert isinstance(Record1.relations.nested_languages, PKNestedListRelation)

    # Empty initialization
    record = Record1({})
    assert record == {}
    assert isinstance(record.relations, RelationsMapping)
    assert callable(record.relations.nested_languages)
    assert record.relations.nested_languages() is None

    # Initialize from dictionary data
    record = Record1(
        {
            "nested_languages": [
                [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
                [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
            ]
        }
    )

    _test_iterators(record)

    # Initialize from field data
    record = Record1(
        {"metadata": {"title": "random title"}},
        relations={
            "nested_languages": [
                [str(en_lang.id), str(es_lang.id)],
                [str(fr_lang.id), str(en_lang.id)],
            ]
        },
    )
    _test_inner_values(record)
    _test_iterators(record)

    # Set via record dictionary data
    record = Record1.create({})
    record["nested_languages"] = [
        [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
        [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
    ]

    _test_iterators(record)

    record.commit()  # validates

    _test_iterators(record)

    # fails with non-list types
    record["nested_languages"] = [{"id": str(es_lang.id)}]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # fails with invalid values
    record["nested_languages"] = [[{"id": "invalid"}]]
    with pytest.raises(InvalidRelationValue):
        record.commit()  # fails validation

    # Set via attribute
    record = Record1.create({})
    record.relations.nested_languages = [
        [str(en_lang.id), str(es_lang.id)],
        [str(fr_lang.id), str(en_lang.id)],
    ]
    _test_inner_values(record)
    _test_iterators(record)

    # Update existing value
    record.relations.nested_languages = [
        [str(fr_lang.id), str(en_lang.id)],
    ]
    assert record["nested_languages"][0][0]["id"] == (str(fr_lang.id))
    assert record["nested_languages"][0][1]["id"] == (str(en_lang.id))
    res_iter = record.relations.nested_languages()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Set invalid value
    with pytest.raises(InvalidRelationValue):
        record.relations.nested_languages = "invalid"
    # Check that old value is still there
    assert record["nested_languages"][0][0]["id"] == (str(fr_lang.id))
    assert record["nested_languages"][0][1]["id"] == (str(en_lang.id))
    res_iter = record.relations.nested_languages()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Dereference relation
    record.relations.nested_languages.dereference()
    assert record["nested_languages"] == [
        [  # last updated values are fr + en
            {
                "id": str(fr_lang.id),
                "iso": "fr",
                "@v": str(fr_lang.id) + "::" + str(fr_lang.revision_id),
            },
            {
                "id": str(en_lang.id),
                "iso": "en",
                "@v": str(en_lang.id) + "::" + str(en_lang.revision_id),
            },
        ],
    ]

    # Commit clean dereferenced fields
    record.commit()
    assert record["nested_languages"][0][0]["id"] == (str(fr_lang.id))
    assert record["nested_languages"][0][1]["id"] == (str(en_lang.id))

    # Commit to DB and refetch
    db.session.commit()
    record = Record1.get_record(record.id)
    assert record["nested_languages"][0][0]["id"] == (str(fr_lang.id))
    assert record["nested_languages"][0][1]["id"] == (str(en_lang.id))
    res_iter = record.relations.nested_languages()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)

    # Clear field
    record.relations.nested_languages = None
    assert "nested_languages" not in record
    assert record.relations.nested_languages() is None
    record.commit()


def test_relations_field_pk_relation_with_value_check(testapp, db, languages):
    """RelationsField tests for PKRelation with value_check param."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]
    oe_lang = languages["oe"]

    # Incorrect record definition with 1 non-list value to check
    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key="language",
                keys=["iso"],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity="English")),
            )
        )

    # Correct record definition with a list of values to check for
    class Record2(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key="language",
                keys=["iso"],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity=["English", "French"])),
            )
        )

    # Wrong record definition, value_check must contain existing fields in the
    # vocabulary value
    class Record3(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key="language",
                keys=["iso"],
                record_cls=Language,
                value_check=dict(wrong_value=dict(ethnicity=["English"])),
            )
        )

    # Correct record definition with multiple values to check
    class Record4(Record, SystemFieldsMixin):
        relations = RelationsField(
            language=PKRelation(
                key="language",
                keys=["iso"],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity=["English"]), iso=["oe"]),
            )
        )

    record = Record1.create({})
    record["language"] = {"id": str(en_lang.id)}
    # Fails because value_check must be a list
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record2.create({})
    record["language"] = {"id": str(es_lang.id)}
    # Fails because only ethnicity accepted is English and French
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record["language"] = {"id": str(fr_lang.id)}
    record.commit()  # validates
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == fr_lang

    record = Record3.create({})
    record["language"] = {"id": str(en_lang.id)}
    # Fails because wrong_value is not a field of the language vocabulary
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record4.create({})
    record["language"] = {"id": str(en_lang.id)}
    # Fails because only records with English ethnicity and 'oe' as iso are
    # accepted
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record["language"] = {"id": str(oe_lang.id)}
    record.commit()  # validates
    res = record.relations.language()
    assert isinstance(res, Language)
    assert res == oe_lang


def test_relations_field_pk_list_relation_with_value_check(testapp, db, languages):
    """RelationsField tests for PKListRelation with value_check param."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]
    oe_lang = languages["oe"]

    # Incorrect record definition with 1 non-list value to check
    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key="languages",
                keys=["iso", "information.ethnicity"],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity="English")),
            )
        )

    # Correct record definition with a list of values to check for
    class Record2(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key="languages",
                keys=["iso", "information.ethnicity"],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity=["English", "French"])),
            )
        )

    # Wrong record definition, value_check must contain existing fields in the
    # vocabulary value
    class Record3(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key="languages",
                keys=["iso", "information.ethnicity"],
                record_cls=Language,
                value_check=dict(wrong_value=dict(ethnicity=["English"])),
            )
        )

    # Correct record definition with multiple values to check
    class Record4(Record, SystemFieldsMixin):
        relations = RelationsField(
            languages=PKListRelation(
                key="languages",
                keys=["iso", "information.ethnicity"],
                record_cls=Language,
                value_check=dict(information=dict(ethnicity=["English"]), iso=["oe"]),
            )
        )

    record = Record1.create({})
    record["languages"] = [{"id": str(fr_lang.id)}]
    # Fails because value_check must be a list
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record2.create({})
    record["languages"] = [{"id": str(es_lang.id)}, {"id": str(en_lang.id)}]
    # Fails because only ethnicity accepted is English and French
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record["languages"] = [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}]
    record.commit()  # validates
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == fr_lang

    record = Record3.create({})
    record["languages"] = [{"id": str(en_lang.id)}]
    # Fails because wrong_value is not a field of the language vocabulary
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record4.create({})
    record["languages"] = [{"id": str(en_lang.id)}]
    # Fails because only records with English ethnicity and 'oe' as iso are
    # accepted
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record["languages"] = [{"id": str(oe_lang.id)}]
    record.commit()  # validates
    res_iter = record.relations.languages()
    assert isinstance(res_iter, Iterable)
    res = next(res_iter)
    assert res == oe_lang


def test_relations_field_pk_nested_list_of_obj_w_related_field_w_value_check(
    testapp, db, languages
):
    """RelationsField tests for PKListRelation with a list of objects and
    with the value_check param."""

    Language, languages = languages
    en_lang = languages["en"]
    fr_lang = languages["fr"]
    es_lang = languages["es"]
    oe_lang = languages["oe"]

    # Incorrect record definition with 1 non-list value to check
    class Record1(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key="nested_array_of_objects",
                keys=["iso"],
                record_cls=Language,
                relation_field="languages",
                value_check=dict(information=dict(ethnicity="English")),
            ),
        )

    # Correct record definition with a list of values to check for
    class Record2(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key="nested_array_of_objects",
                keys=["iso"],
                record_cls=Language,
                relation_field="languages",
                value_check=dict(information=dict(ethnicity=["English", "French"])),
            ),
        )

    # Wrong record definition, value_check must contain existing fields in the
    # vocabulary value
    class Record3(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key="nested_array_of_objects",
                keys=["iso"],
                record_cls=Language,
                relation_field="languages",
                value_check=dict(wrong_value=dict(ethnicity=["English"])),
            ),
        )

    # Correct record definition with multiple values to check
    class Record4(Record, SystemFieldsMixin):
        relations = RelationsField(
            nested_array_of_objects=PKNestedListRelation(
                key="nested_array_of_objects",
                keys=["iso"],
                record_cls=Language,
                relation_field="languages",
                value_check=dict(information=dict(ethnicity=["English"]), iso=["oe"]),
            ),
        )

    record = Record1.create({})
    record["nested_array_of_objects"] = [
        {
            "languages": [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
            "field1": "string",
        },
        {
            "languages": [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
            "field2": "string2",
        },
    ]
    # Fails because value_check must be a list
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record2.create({})
    record["nested_array_of_objects"] = [
        {
            "languages": [{"id": str(en_lang.id)}, {"id": str(es_lang.id)}],
            "field1": "string",
        },
        {
            "languages": [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
            "field2": "string2",
        },
    ]
    # Fails because only ethnicity accepted is English and French
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record["nested_array_of_objects"] = [
        {
            "languages": [{"id": str(en_lang.id)}, {"id": str(fr_lang.id)}],
            "field1": "string",
        },
        {
            "languages": [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
            "field2": "string2",
        },
    ]

    record.commit()  # validates
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == en_lang
    res = next(res_inner_iter)
    assert res == fr_lang
    with pytest.raises(StopIteration):  # finished first iterator
        res = next(res_inner_iter)
    # second inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == fr_lang
    res = next(res_inner_iter)
    assert res == en_lang
    with pytest.raises(StopIteration):  # finished second iterator
        res = next(res_inner_iter)

    record = Record3.create({})
    record["nested_array_of_objects"] = [
        {
            "languages": [{"id": str(en_lang.id)}, {"id": str(fr_lang.id)}],
            "field1": "string",
        },
        {
            "languages": [{"id": str(fr_lang.id)}, {"id": str(en_lang.id)}],
            "field2": "string2",
        },
    ]
    # Fails because wrong_value is not a field of the language vocabulary
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record = Record4.create({})
    record["nested_array_of_objects"] = [
        {"languages": [{"id": str(en_lang.id)}], "field1": "string"}
    ]
    # Fails because only ethnicity accepted is English
    with pytest.raises(InvalidCheckValue):
        record.commit()  # validates

    record["nested_array_of_objects"] = [
        {"languages": [{"id": str(oe_lang.id)}], "field1": "string"}
    ]
    res_iter = record.relations.nested_array_of_objects()
    assert isinstance(res_iter, Iterable)
    # first inner list
    res_inner_iter = next(res_iter)
    assert isinstance(res_inner_iter, Iterable)
    res = next(res_inner_iter)
    assert res == oe_lang
    with pytest.raises(StopIteration):  # finished first iterator
        next(res_inner_iter)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Dictionary utilities."""

from copy import deepcopy


def clear_none(d):
    """Clear None values and empty dicts from a dict."""
    del_keys = []
    for k, v in d.items():
        if v is None:
            del_keys.append(k)
        elif isinstance(v, dict):
            clear_none(v)
            if v == {}:
                del_keys.append(k)
        elif isinstance(v, list):
            clear_none_list(v)
            if v == []:
                del_keys.append(k)

    # Delete the keys (cannot be done during the dict iteration)
    for k in del_keys:
        del d[k]


def clear_none_list(ls):
    """Clear values from a list (in-place)."""
    del_idx = []
    for i, v in enumerate(ls):
        if v is None:
            del_idx.append(i)
        elif isinstance(v, list):
            clear_none_list(v)
            if v == []:
                del_idx.append(i)
        elif isinstance(v, dict):
            clear_none(v)
            if v == {}:
                del_idx.append(i)

    # Delete the keys (reverse so index stays stable).
    for i in reversed(del_idx):
        del ls[i]


def parse_lookup_key(lookup_key):
    """Parse a lookup key."""
    if not lookup_key:
        raise KeyError("No lookup key specified")

    # Parse the list of keys
    if isinstance(lookup_key, str):
        keys = lookup_key.split(".")
    elif isinstance(lookup_key, list):
        keys = lookup_key
    else:
        raise TypeError("lookup must be string or list")

    return keys


def dict_lookup(source, lookup_key, parent=False):
    """Make a lookup into a dict based on a dot notation.

    Examples of the supported dot notation:

    - ``'a'`` - Equivalent to ``source['a']``
    - ``'a.b'`` - Equivalent to ``source['a']['b']``
    - ``'a.b.0'`` - Equivalent to ``source['a']['b'][0]`` (for lists)

    List notation is also supported:

    - `['a']``
    - ``['a','b']``
    - ``['a','b', 0]``

    :param source: The dictionary object to perform the lookup in.
    :param parent: If parent argument is True, returns the parent node of
                   matched object.
    :param lookup_key: A string using dot notation, or a list of keys.
    """
    # Copied from dictdiffer (CERN contributed part) and slightly modified.
    keys = parse_lookup_key(lookup_key)

    if parent:
        keys = keys[:-1]

    # Lookup the key
    value = source
    for key in keys:
        try:
            if isinstance(value, list):
                key = int(key)
            value = value[key]
        except (TypeError, IndexError, ValueError) as exc:
            raise KeyError(lookup_key) from exc
    return value


def dict_set(source, key, value):
    """Set a value into a dict via a dot-notated key.

    This also creates missing key "paths".

    Examples of the supported dot notation:

    - ``'a'`` - Equivalent to ``source['a'] = value``
    - ``'a.b'`` - Equivalent to ``source['a']['b'] = value``
    - ``'a.b.0'`` - Equivalent to ``source['a']['b'][0] = value`` (for lists)

    List notation is also supported:

    - `['a']``
    - ``['a','b']``
    - ``['a','b', 0]``

    :param source: The dictionary object to set the value in.
    :param key: A string using dot notation, or a list of keys.
    :param value: The value to be set.
    """
    keys = parse_lookup_key(key)
    parent = source
    for key in keys[:-1]:
        if isinstance(key, int):
            parent = parent[key]
        else:
            parent = parent.setdefault(key, {})
    parent[keys[-1]] = value


def dict_merge(dest, source):
    """Merges source into dest.

    It does not merge arrays of dicts.
    """
    for key in source:
        if key in dest:
            if isinstance(dest[key], dict) and isinstance(source[key], dict):
                dict_merge(dest[key], source[key])
        else:
            dest[key] = source[key]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2022 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field."""

from werkzeug.utils import cached_property

from ..base import SystemField
from .mapping import RelationsMapping
from .relations import RelationBase


#
# System field
#
class RelationsField(SystemField):
    """Relations field for connections to external entities."""

    def __init__(self, **fields):
        """Initialize the field."""
        super().__init__()
        assert all(isinstance(f, RelationBase) for f in fields.values())
        self._original_fields = fields

    def __getattr__(self, name):
        """Get a field definition."""
        if name in self._fields:
            return self._fields[name]
        raise AttributeError

    def __iter__(self):
        """Iterate over the configured fields."""
        return iter(getattr(self, f) for f in self._fields)

    def __contains__(self, name):
        """Return if a field exists in the configured fields."""
        return name in self._fields

    #
    # Properties
    #
    @cached_property
    def _fields(self):
        """Get the fields."""
        return self._original_fields

    #
    # Helpers
    #
    def obj(self, instance):
        """Get the relations object."""
        # Check cache
        obj = self._get_cache(instance)
        if obj:
            return obj
        obj = RelationsMapping(record=instance, fields=self._fields)
        self._set_cache(instance, obj)
        return obj

    #
    # Data descriptor
    #
    def __get__(self, record, owner=None):
        """Accessing the attribute."""
        # Class access
        if record is None:
            return self
        return self.obj(record)

    def __set__(self, instance, values):
        """Setting the attribute."""
        obj = self.obj(instance)
        for k, v in values.items():
            setattr(obj, k, v)

    #
    # Record extension
    #
    def pre_commit(self, record):
        """Initialise the model field."""
        self.obj(record).validate()
        self.obj(record).clean()


class MultiRelationsField(RelationsField):
    """Relations field for connections to external entities.

    It allows to define nested relation fields. For example:

    .. code-block:: python

        class Record:

            relations = MultiRelationsField(
                field_one=PIDListRelation(
                    "metadata.field_one",
                    ...
                ),
                inner=RelationsField(
                    inner_field=PIDListRelation(
                        "metadata.inner_field",
                    ...
                    ),
                )
            )
    """

    def __init__(self, **fields):
        """Initialize the field.

        The nested RelationFields will be flattened to the root.
        In the example above, the relations field has a field (field_one)
        and a nested RelationsField with a field (inner_field). However,
        both of them will be accessed through the relations field.

        .. code-block:: python

            relations.field_one
            relations.inner_field  # correct
            relations.inner.inner_field  # incorrect
        """
        super().__init__()  # no fields passed since validation happens in this class
        assert all(
            isinstance(f, RelationBase) or isinstance(f, RelationsField)
            for f in fields.values()
        )

        self._original_fields = fields
        self._relation_fields = set()

    @cached_property
    def _fields(self):
        """Mutates self._fields to include all nested fields."""
        fields = {}
        self._relation_fields = set()
        for key, field in self._original_fields.items():
            if isinstance(field, RelationBase):
                fields[key] = field
            elif isinstance(field, RelationsField):
                self._relation_fields.add(key)
                for inner_name, inner_field in field._fields.items():
                    if inner_name not in fields:
                        fields[inner_name] = inner_field
        return fields

    #
    # Data descriptor
    #
    def __set__(self, instance, values):
        """Setting the attribute."""
        obj = self.obj(instance)
        for k, v in values.items():
            if k in self._relation_fields:
                for kk, vv in v.items():
                    setattr(obj, kk, vv)
            else:
                setattr(obj, k, v)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records."""

import pytest
from invenio_db.utils import drop_alembic_version_table


@pytest.mark.skip(reason="Caused by mergepoint")
def test_alembic(testapp, db):
    """Test alembic recipes."""
    ext = testapp.extensions["invenio-db"]

    if db.engine.name == "sqlite":
        raise pytest.skip("Upgrades are not supported on SQLite.")

    assert not ext.alembic.compare_metadata()
    db.drop_all()
    drop_alembic_version_table()
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()
    ext.alembic.stamp()
    ext.alembic.downgrade(target="96e796392533")
    ext.alembic.upgrade()

    assert not ext.alembic.compare_metadata()
    drop_alembic_version_table()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test Invenio Records."""

from copy import copy
from datetime import date, timedelta

import pytest
from jsonschema.exceptions import ValidationError

from invenio_records import Record


@pytest.fixture()
def encoder():
    """Return a simple encoder."""

    class CustomEncoder:
        @classmethod
        def encode(cls, data):
            if "date" in data.get("nested", {}):
                data["nested"]["date"] = data["nested"]["date"].isoformat()
            return data

        @classmethod
        def decode(cls, data):
            if "date" in data.get("nested", {}):
                datestr = data["nested"]["date"]
                data["nested"]["date"] = date(
                    int(datestr[0:4]),
                    int(datestr[5:7]),
                    int(datestr[8:10]),
                )
            return data

    return CustomEncoder


@pytest.fixture()
def schema():
    """A test schema."""
    return {
        "type": "object",
        "properties": {
            "title": {"type": "string"},
            "nested": {"type": "object", "properties": {"date": {"type": "string"}}},
        },
    }


@pytest.fixture()
def testdate():
    """A test date."""
    return date(2020, 9, 3)


@pytest.fixture()
def testdatestr(testdate):
    """Test date as a string."""
    return testdate.isoformat()


@pytest.fixture()
def record_cls(encoder):
    """Mock the record class with an encoder."""
    Record.model_cls.encoder = encoder
    yield Record
    Record.model_cls.encoder = None


def test_noencoding(testapp, db, record_cls):
    """Simple sanity check."""
    data = {"title": "Title"}
    rec = record_cls.create(data)
    assert dict(rec) == rec.model.json
    assert rec.model.data == rec.model.json


def test_encoding(testapp, db, record_cls, testdate, testdatestr):
    """Test encoding/decoding in custom classes."""
    # Create a record with a Python date inside (record_cls has an encoder
    # which will encode/decode the "date" field if present).
    data = {"title": "Title", "nested": {"date": testdate}}
    rec = record_cls.create(data)
    db.session.commit()

    # Record and data should still hold a python date
    assert data["nested"]["date"] == testdate
    assert rec["nested"]["date"] == testdate

    # The underlying model JSON field should have a string instead.
    assert rec.model.json["nested"]["date"] == testdatestr

    # Decoding the data: Get the record again
    rec = record_cls.get_record(rec.id)

    # Record and data should still hold a python date
    assert rec["nested"]["date"] == testdate

    # The underlying model JSON field should have a string instead.
    assert rec.model.json["nested"]["date"] == testdatestr


def test_encoding_with_versioning(testapp, database, record_cls, testdate):
    """Test reverting a revision."""
    db = database
    # Create a record and a new revision
    rec = record_cls.create({"title": "Title", "nested": {"date": testdate}})
    db.session.commit()

    # Change record (making a new revision)
    newdate = rec["nested"]["date"] + timedelta(days=1)
    rec["nested"]["date"] = newdate
    rec.commit()
    db.session.commit()

    # Record and model should have the new date.
    assert rec["nested"]["date"] == newdate
    assert rec.model.json["nested"]["date"] == newdate.isoformat()

    # Now revert the record and check against initial date.
    rec = rec.revert(0)
    assert rec["nested"]["date"] == testdate
    assert rec.model.json["nested"]["date"] == testdate.isoformat()


def test_encoding_with_schema(testapp, database, record_cls, testdate, schema):
    """Test that schema validation works on JSON, not on dict."""
    db = database

    # Create a record
    rec = record_cls.create(
        {
            "$schema": schema,
            "title": "Title",
            "nested": {"date": testdate},
        }
    )
    db.session.commit()
    # No assertion: because JSONSchema validation will raise an Exception
    # if it fails to operate the date

    # Try on record update as well (since commit() also validates)
    newdate = rec["nested"]["date"] + timedelta(days=1)
    rec["nested"]["date"] = newdate
    rec.commit()
    db.session.commit()
    # No assertion: because JSONSchema validation will raise an Exception
    # if it fails to operate the date


def test_encoding_with_schema_fail(testapp, database, testdate, schema):
    """Test that validation fails when not using encoder."""
    db = database

    data = {
        "$schema": schema,
        "title": "Title",
        "nested": {"date": testdate},
    }

    assert pytest.raises(ValidationError, Record.create, data)

    # Create empty record so we can test Record.commit()
    rec = Record.create({})
    db.session.commit()

    rec.update(data)
    assert pytest.raises(ValidationError, rec.commit)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Records-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Persistent identifier system field for record."""

from ..dictutils import clear_none, dict_lookup
from .base import SystemField


class DictField(SystemField):
    """Dictionary field.

    Provides a shortcut for getting/setting a specific key on a record.
    """

    def __init__(self, key=None, clear_none=False, create_if_missing=True):
        """Initialise the dict field.

        :param key: Key to set (dot notation supported).
        :param clear_none: Boolean to control if empty/None values should be
                           removed.
        :param create_if_missing: If a subkey is missing it will be created if
                                  this option is set to true.
        """
        self.clear_none = clear_none
        self.create_if_missing = create_if_missing
        super().__init__(key=key)

    def __get__(self, record, owner=None):
        """Getting the attribute value."""
        if record is None:
            return self
        return self.get_dictkey(record)

    def __set__(self, record, value):
        """Setting a new value."""
        if self.clear_none:
            clear_none(value)
        self.set_dictkey(record, value, create_if_missing=self.create_if_missing)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Constant system field."""

from ..dictutils import dict_lookup
from .base import SystemField


class ModelField(SystemField):
    """Model field for providing get and set access on a model field."""

    def __init__(
        self, model_field_name=None, dump=True, dump_key=None, dump_type=None, **kwargs
    ):
        """Initialize the field.

        :param model_field_name: Name of field on the database model.
        :param dump: Set to false to not dump the field.
        :param dump_key: The dictionary key to use in dumps.
        :param dump_type: The data type used to determine how to serialize the
            model field.
        """
        self._model_field_name = model_field_name
        self.dump = dump
        self._dump_key = dump_key
        self._dump_type = dump_type
        super().__init__(**kwargs)

    #
    # Helpers
    #
    @property
    def model_field_name(self):
        """The name of the SQLAlchemy field on the model.

        Defaults to the attribute name used on the class.
        """
        return self._model_field_name or self.attr_name

    @property
    def dump_key(self):
        """The dictionary key to use in dump output.

        Note, it's up to the dumper to choose if it respects this name.
        The name defaults to the model field name.
        """
        return self._dump_key or self.model_field_name

    @property
    def dump_type(self):
        """The data type used to determine how to serialize the model field.

        Defaults to none, meaning the dumper will determine how to dump it.
        """
        return self._dump_type

    def _set(self, model, value):
        """Internal method to set value on the model's field."""
        setattr(model, self.model_field_name, value)

    #
    # Data descriptor
    #
    def __get__(self, record, owner=None):
        """Accessing the attribute."""
        # Class access
        if record is None:
            return self
        # Instance access
        try:
            return getattr(record.model, self.model_field_name)
        except AttributeError:
            return None

    def __set__(self, instance, value):
        """Accessing the attribute."""
        self._set(instance.model, value)

    #
    # Record extension
    #
    def post_init(self, record, data, model=None, field_data=None):
        """Initialise the model field."""
        if field_data is not None:
            self._set(model, field_data)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test for model field.

This tests needs to live in it's own file to have a clean database session.
"""


from datetime import datetime

import pytest

from invenio_records.api import Record
from invenio_records.dumpers import SearchDumper
from invenio_records.models import RecordMetadataBase
from invenio_records.systemfields import ModelField, SystemFieldsMixin


@pytest.fixture(scope="module")
def Record1Metadata(database):
    """."""
    db = database

    class Record1Metadata(db.Model, RecordMetadataBase):
        __tablename__ = "record1_metadata"

        expires_at = db.Column(db.DateTime())

    Record1Metadata.__table__.create(db.engine)
    return Record1Metadata


def test_model_field(testapp, database, Record1Metadata):
    """Test model field with clearing none/empty values."""
    db = database

    class Record1(Record, SystemFieldsMixin):
        model_cls = Record1Metadata
        dumper = SearchDumper()
        # Don't do this at home (two system fields on the same model field):
        expires_at = ModelField()
        expires = ModelField("expires_at", dump=False)

    dt = datetime(2020, 9, 3, 0, 0)
    dt2 = datetime(2020, 9, 4, 0, 0)

    # Field itself
    assert isinstance(Record1.expires_at, ModelField)

    # Test creating a record with model
    record = Record1.create({}, expires_at=dt)
    db.session.commit()
    assert record.expires_at == dt
    assert record.model.expires_at == dt

    # Test that another field name can be used:
    record = Record1.create({}, expires=dt)
    db.session.commit()
    assert record.expires == dt
    assert record.model.expires_at == dt

    # Test getting a record
    record = Record1.get_record(record.id)
    assert record.expires_at == dt
    assert record.model.expires_at == dt

    # Test assignment
    dt2 = datetime(2020, 9, 4, 0, 0)
    record.expires_at = dt2
    db.session.commit()
    assert record.expires_at == dt2
    assert record.model.expires_at == dt2

    # Test creating a record without model
    record = Record1({})
    assert record.expires_at is None
    assert record.model is None

    # Creation without model, but with arg will raise attribute error.
    pytest.raises(AttributeError, Record1, {}, expires_at=dt2)
    # Creation without model and without arg is ok.
    assert Record1({}) == {}

    # Test dumping and loading
    record = Record1.create({}, expires_at=dt)
    dump = record.dumps()
    loaded_record = record.loads(dump)
    assert loaded_record.expires_at == record.expires_at

    # Test dumping with None
    record = Record1.create({})
    dump = record.dumps()
    loaded_record = record.loads(dump)
    assert loaded_record.expires_at is None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Records is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Related model field.

The related model field serializes/dumps a related object into the record JSON
dictionary (basically denormalization). This way, the object can be recreated
directly from the record JSON dictionary instead of querying the database.

This is useful for instance during search queries to avoid hitting the
database. In addition you can subclass this field to provide life-cycle hooks
on the record. For instance a PIDField, could create the related persistent
identifier.
"""

from invenio_db import db
from sqlalchemy import inspect

from .base import SystemField, SystemFieldContext


class RelatedModelFieldContext(SystemFieldContext):
    """Context for RelatedModelField.

    This class implements the class-level methods available on a
    RelatedModelField. I.e. when you access the field through the class, for
    instance:

    .. code-block:: python

        Record.myattr.session_merge(record)
    """

    def session_merge(self, record):
        """Merge the PID to the session if not persistent."""
        obj = self.field.obj(record)
        if not inspect(obj).persistent:
            obj = db.session.merge(obj)
            self.field._set_cache(record, obj)


class RelatedModelField(SystemField):
    """Related model system field."""

    def __init__(
        self, model, key=None, required=False, load=None, dump=None, context_cls=None
    ):
        """Initialize the field.

        :param model: Related SQLAlchemy model.
        :param key: Name of key in the record to serialize the related object
            under.
        :param required: Flag to determine if a related object is required on
            record commit time.
        :param load: Callable to load the related object from a JSON object.
        :param dump: Callable to dump the related object as a JSON object.
        :param context_cls: The context class is used to provide additional
            methods on the field itself.
        """
        self._model = model
        self._required = required
        self._load = load or model.load_obj
        self._dump = dump or model.dump_obj
        self._context_cls = context_cls or RelatedModelFieldContext
        super().__init__(key=key)

    #
    # Life-cycle hooks
    #
    def pre_commit(self, record):
        """Called before a record is committed."""
        # Make sure we serialize/dump the related objet on record.commit() time
        # as it might have changed.
        related_obj = getattr(record, self.attr_name)
        if related_obj is not None:
            self.set_obj(record, related_obj)
        elif self._required:
            raise RuntimeError("You must provide a related object.")

    #
    # Helpers
    #
    def obj(self, record):
        """Get the related object.

        Uses a cached object if it exists.

        IMPORTANT: By default, if the object is loaded from the record JSON
        object instead of from the database model, it is NOT added to the
        database session. Thus, the related object will be in a transient state
        instead of persistent state. This is useful for instance in search
        queries to avoid hitting the database, however if you need to make
        operations on it you should add it to the session using:

        .. code-block::

            Record.myattr.session_merge(record)
        """
        # Check cache
        obj = self._get_cache(record)
        if obj:
            return obj

        obj = self._load(self, record)
        if obj:
            # Cache object
            self._set_cache(record, obj)
            return obj
        return None

    def set_obj(self, record, obj):
        """Set the object."""
        assert isinstance(obj, self._model)

        # Store data values on the attribute name (e.g. 'type') using dump
        # method provided by either to the field or a function on the related
        # object.
        self._dump(self, record, obj)

        # Cache object
        self._set_cache(record, obj)

    #
    # Data descriptor methods (i.e. attribute access)
    #
    def __get__(self, record, owner=None):
        """Get the persistent identifier."""
        if record is None:
            return self._context_cls(self, owner)
        return self.obj(record)

    def __set__(self, record, pid):
        """Set persistent identifier on record."""
        self.set_obj(record, pid)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 RERO.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test for system fields."""

import pytest

from invenio_records.api import Record
from invenio_records.dumpers import SearchDumper
from invenio_records.systemfields import (
    ConstantField,
    DictField,
    SystemField,
    SystemFieldsMixin,
)


#
# Fixtures
#
@pytest.fixture()
def testschema():
    """The JSONSchema used for tests."""
    return {
        "type": "object",
        "properties": {
            "title": {"type": "string"},
        },
        "required": ["title"],
    }


@pytest.fixture()
def SystemRecord():
    """System record is a base class."""

    class SystemRecord(Record, SystemFieldsMixin):
        # Only defined in SystemRecord:
        test = ConstantField("test", "test")
        # Defined in both SystemRecord and MyRecord:
        base = ConstantField("base", "systemrecord")

    return SystemRecord


@pytest.fixture()
def MyRecord(SystemRecord, testschema):
    """My record class inheriting from SystemRecord."""

    class MyRecord(SystemRecord):
        # Defined in only MyRecord
        schema = ConstantField("$schema", testschema)
        # Defined in both SystemRecord and MyRecord
        base = ConstantField("base", "myrecord")

    return MyRecord


@pytest.fixture()
def record(MyRecord, db):
    """Create a record instance of MyRecord."""
    return MyRecord.create({"title": "test"})


@pytest.fixture()
def sysrecord(SystemRecord, db):
    """Create a record instance of SystemRecord."""
    return SystemRecord.create({"title": "test"})


@pytest.fixture()
def ExtensionRecord():
    """Create an ExtensionRecord class to test extensions being called."""

    class ExtensionField(SystemField):
        called = []

        def pre_init(self, record, data, model=None, **kwargs):
            self.called.append("pre_init")

        def post_init(self, record, data, model=None, field_data=None):
            self.called.append("post_init")

        def pre_dump(self, record, dumper=None):
            self.called.append("pre_dump")

        def post_load(self, record, loader=None):
            self.called.append("post_load")

        def pre_create(self, record):
            self.called.append("pre_create")

        def post_create(self, record):
            self.called.append("post_create")

        def pre_commit(self, record):
            self.called.append("pre_commit")

        def post_commit(self, record):
            self.called.append("post_commit")

        def pre_delete(self, record, force=False):
            self.called.append("pre_delete")

        def post_delete(self, record, force=False):
            self.called.append("post_delete")

        def pre_revert(self, record, revision):
            self.called.append("pre_revert")

        def post_revert(self, new_record, revision):
            self.called.append("post_revert")

    class ExtensionRecord(Record, SystemFieldsMixin):
        dumper = SearchDumper()
        ext = ExtensionField()

    return ExtensionRecord


#
# Tests
#
def test_constant_field(testapp, record, MyRecord, testschema):
    """Test the constant field."""
    # Test that constant field value was set, and is accessible
    assert record["$schema"] == testschema
    assert record.schema == testschema

    # Test that constant fields cannot have values assigned.
    try:
        record.schema = testschema
        assert False
    except AttributeError:
        pass

    # Explicit remove the set value ($schema) after the record creation.
    record = MyRecord({"title": "test"})
    del record["$schema"]

    # Now test, that accessing the field returns None.
    assert record.schema is None


def test_constant_field_deletion(testapp, db, record, MyRecord, testschema):
    """Test the constant field."""
    # Test that constant field value was set, and is accessible
    db.session.commit()
    record.delete()
    db.session.commit()

    record = MyRecord.get_record(record.id, with_deleted=True)
    assert record.schema is None


def test_field_overwriting(testapp, record, sysrecord):
    """Test that field overwriting."""
    # field 'base' is defined in both classes, thus MyRecord overwrites
    # SystemRecord.
    assert record.base == "myrecord"
    assert sysrecord.base == "systemrecord"


def test_field_inheritance(testapp, record, sysrecord):
    """Test that field inheritance."""
    # field 'test' is defined only in base class, but available on both.
    assert record.test == "test"
    assert sysrecord.test == "test"


def test_field_class_access(testapp, MyRecord):
    """Test that class access is returning the field."""
    assert isinstance(MyRecord.schema, ConstantField)
    assert isinstance(MyRecord.test, ConstantField)
    assert isinstance(MyRecord.base, ConstantField)


def test_attrname_injection(testapp, MyRecord):
    """Test that class access is returning the field."""
    assert MyRecord.schema.attr_name == "schema"
    assert MyRecord.test.attr_name == "test"
    assert MyRecord.base.attr_name == "base"


def test_systemfields_mro(testapp):
    """Test overwriting of system fields according to MRO."""

    class A(Record, SystemFieldsMixin):
        test = ConstantField("test", "a")

    class B(A):
        test = ConstantField("test", "b")

    class C(A):
        pass

    class D(B, C):
        test = ConstantField("test", "d")

    class E(B, C):
        pass

    class F(C):
        pass

    class G(C, B):
        pass

    # A defines field itself
    assert A({}).test == "a"
    # B inherits from A  (overwrites field)
    assert B({}).test == "b"
    # C inherits from A
    assert C({}).test == "a"
    # D inherits from B and C (overwrites field)
    assert D({}).test == "d"
    # E inherits from B and C
    assert E({}).test == "b"
    # F inherits from C which inherits from A.
    assert F({}).test == "a"
    # G inherits from C (which inherits from A) and B.
    assert G({}).test == "b"


def test_extension_pre_init(testapp, db, ExtensionRecord):
    """Test pre init hook."""
    rec = ExtensionRecord({})
    assert ExtensionRecord.ext.called == ["pre_init", "post_init"]


def test_extension_post_create(testapp, db, ExtensionRecord):
    """Test post create hook."""
    rec = ExtensionRecord.create({})
    assert ExtensionRecord.ext.called == [
        "pre_init",
        "post_init",
        "pre_create",
        "post_create",
    ]


def test_extension_pre_dump(testapp, db, ExtensionRecord):
    """Test pre dump hook."""
    rec = ExtensionRecord({}).dumps()
    assert ExtensionRecord.ext.called == ["pre_init", "post_init", "pre_dump"]


def test_extension_post_load(testapp, db, ExtensionRecord):
    """Test post load hook."""
    dump = ExtensionRecord({}).dumps()
    rec = ExtensionRecord.loads(dump)
    assert ExtensionRecord.ext.called == [
        "pre_init",
        "post_init",
        "pre_dump",
        "pre_init",
        "post_init",
        "post_load",
    ]


def test_extension_commit(testapp, db, ExtensionRecord):
    """Test commit hooks."""
    rec = ExtensionRecord.create({})
    rec.commit()
    assert ExtensionRecord.ext.called == [
        "pre_init",
        "post_init",
        "pre_create",
        "post_create",
        "pre_commit",
        "post_commit",
    ]


def test_extension_delete(testapp, db, ExtensionRecord):
    """Test pre/post delete hook."""
    rec = ExtensionRecord.create({})
    db.session.commit()
    rec.delete()
    assert ExtensionRecord.ext.called == [
        "pre_init",
        "post_init",
        "pre_create",
        "post_create",
        "pre_delete",
        "post_delete",
    ]


def test_extension_revert(testapp, database, ExtensionRecord):
    """Test pre/post revert hook."""
    db = database
    rec = ExtensionRecord.create({})
    db.session.commit()
    rec.commit()
    db.session.commit()
    assert rec.revision_id == 1
    rec.revert(0)
    assert ExtensionRecord.ext.called == [
        "pre_init",
        "post_init",
        "pre_create",
        "post_create",
        "pre_commit",
        "post_commit",
        "pre_revert",
        "pre_init",
        "post_init",
        "post_revert",
    ]


def test_base_systemfield_base(testapp):
    """Test default implementation of system field."""

    class TestRecord(Record, SystemFieldsMixin):
        field = SystemField()

    # Test to please the test coverage gods
    assert pytest.raises(AttributeError, getattr, TestRecord({}), "field")
    # A core system field doesn't support assignment, so you can't pass args
    # in constructor as well
    assert pytest.raises(AttributeError, TestRecord, {}, field={})


def test_systemfield_initialization(testapp):
    """Test default implementation of system field."""

    class TestField(SystemField):
        def __set__(self, instance, value):
            instance["arg_value"] = value

    class TestRecord(Record, SystemFieldsMixin):
        afield = TestField()

    # Init method
    assert TestRecord({}, afield="testval")["arg_value"] == "testval"

    # Create method
    record = TestRecord.create({}, afield="testval")
    assert record["arg_value"] == "testval"


def test_default_key(testapp):
    """Test default key name."""

    class TestRecord(Record, SystemFieldsMixin):
        afield = ConstantField(value="test")

    assert TestRecord({})["afield"] == "test"


#
# DictField
#
def test_dict_field_simple():
    """Simple tests for the DictField."""

    class Record1(Record, SystemFieldsMixin):
        metadata = DictField("metadata")

    assert isinstance(Record1.metadata, DictField)

    # Creation
    assert Record1({}, metadata={"title": 1}) == {"metadata": {"title": 1}}
    # Access key exists
    assert Record1({"metadata": {"title": 1}}).metadata == {"title": 1}
    # Access key doesn't exists
    assert Record1({}).metadata is None

    # Double tap - metadata wins over default data.
    assert Record1({"metadata": {"title": 2}, "a": "b"}, metadata={"title": 1}) == {
        "metadata": {"title": 1},
        "a": "b",
    }


def test_dictfield_default_key(testapp):
    """Test default key name."""

    class TestRecord(Record, SystemFieldsMixin):
        afield = DictField()

    assert TestRecord({}, afield={"test": 1})["afield"] == {"test": 1}


def test_dict_field_complex_key():
    """More complex tests for the DictField."""

    class Record1(Record, SystemFieldsMixin):
        metadata = DictField("metadata.a.b")

    # Creation
    record = Record1({}, metadata={"title": 1})
    assert record == {"metadata": {"a": {"b": {"title": 1}}}}
    # Access key exists
    assert record.metadata == {"title": 1}
    # Access key doesn't exists
    assert Record1({}).metadata is None

    # Cannot handle lists
    pytest.raises(KeyError, Record1, {"metadata": []}, metadata={"title": 1})
    pytest.raises(KeyError, Record1, {"metadata": {"a": []}}, metadata={"title": 1})


def test_dict_field_create_if_missing():
    """More complex tests for the DictField."""

    class Record1(Record, SystemFieldsMixin):
        metadata = DictField("metadata.a.b", create_if_missing=False)

    # If subkey doesn't exists, it fails with create_if_missing=False
    pytest.raises(KeyError, Record1, {}, metadata={"title": 1})
    pytest.raises(KeyError, Record1, {"metadata": []}, metadata={"title": 1})
    # Subkey exists, all good:
    assert Record1({"metadata": {"a": {}}}, metadata={"title": 1})


def test_dict_field_clear_none():
    """Test dict field with clearing none/empty values."""

    class Record1(Record, SystemFieldsMixin):
        metadata = DictField("metadata", clear_none=True)

    record = Record1({"a": None}, metadata={"a": None, "title": 1, "b": {"c": None}})
    assert record == {"a": None, "metadata": {"title": 1}}

    record = Record1({}, metadata={"a": None})
    assert record == {"metadata": {}}


def test_dict_field_deleted(testapp, database):
    """Test dict field with clearing none/empty values."""
    db = database

    class Record1(Record, SystemFieldsMixin):
        metadata = DictField()

    record = Record1.create({}, metadata={"title": 1})
    db.session.commit()
    record.delete()
    db.session.commit()

    # Loading a deleted record
    record = Record1.get_record(record.id, with_deleted=True)
    assert record.metadata is None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Tests Invenio-Records JSONSchema ref resolver."""

import pytest

from invenio_records.api import Record
from invenio_records.resolver import urljoin_with_custom_scheme


def local_ref_resolver_store_factory():
    """Build local JSONSchema ref resolver store."""
    return {
        "local://authors.json": {
            "$id": "local://authors.json",
            "type": "array",
            "items": {"type": "string"},
        },
        "local://books.json": {
            "$id": "local://books.json",
            "type": "object",
            "definitions": {
                "price": {
                    "type": "string",
                },
            },
            "properties": {
                "title": {"type": "string"},
                "price": {"$ref": "#/definitions/price"},
                "authors": {"$ref": "local://authors.json"},
            },
        },
    }


@pytest.fixture(scope="module")
def app_config(app_config):
    app_config[
        "RECORDS_REFRESOLVER_CLS"
    ] = "invenio_records.resolver.InvenioRefResolver"
    app_config["RECORDS_REFRESOLVER_STORE"] = local_ref_resolver_store_factory()
    return app_config


def test_invenio_refresolver_with_local_store(db):
    """Test InvenioRefResolver with local store and complex JSONSchema."""
    data = {
        "$schema": "local://books.json#",
        "title": "The Lord of the Rings",
        "price": "20",
        "authors": [
            "J. R. R. Tolkien",
        ],
    }
    Record.create(data).commit()
    db.session.commit()


@pytest.mark.parametrize(
    "resolution_scope, scope, expected_output",
    [
        # Custom scheme simple cases
        ("local://books.json", "local://books.json#", "local://books.json"),
        ("local://books.json#", "local://books.json", "local://books.json"),
        ("local://books.json#", "#", "local://books.json"),
        # Custom scheme with internal refs
        (
            "local://books.json",
            "local://books.json#/definitions/price",
            "local://books.json#/definitions/price",
        ),
        (
            "local://books.json",
            "#/definitions/price",
            "local://books.json#/definitions/price",
        ),
    ],
)
def test_urljoin_with_custom_scheme(resolution_scope, scope, expected_output):
    """Test urljoin supporting custom schemas."""
    assert urljoin_with_custom_scheme(resolution_scope, scope) == expected_output



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020-2021 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Relations system field errors."""

from ...errors import RecordsError


class RelationError(RecordsError):
    """Base relation error class."""


class InvalidRelationValue(RelationError):
    """Invalid relation value."""


class InvalidCheckValue(RelationError):
    """Invalid check value."""
